"""
Define pregnancy, deaths, migration, etc.
"""
import numpy as np
import starsim as ss
import sciris as sc
import pandas as pd

ss_float = ss.dtypes.float
ss_int = ss.dtypes.int
_ = None


__all__ = ['Demographics', 'Births', 'Deaths', 'PregnancyPars', 'Pregnancy', 'FetalHealth']


class Demographics(ss.Module):
    """
    A demographic module typically handles births/deaths/migration and takes
    place at the start of the timestep, before networks are updated and before
    any disease modules are executed.
    """
    pass


class Births(Demographics):
    """
    Create births based on rates, rather than based on pregnancy.

    Births are generated using population-level birth rates that can vary
    by year. The number of births per timestep is determined by applying
    the birth rate to the current population size.

    Args:
        birth_rate (float/rate/dataframe): value for birth rate, or birth rate data
        rel_birth (float): constant used to scale all birth rates
        rate_units (float): units for birth rates (default assumes per 1000)
        metadata (dict): dict with data column mappings for birth rate data (if birth_rate is a dataframe)
    """
    def __init__(self, pars=None, birth_rate=_, rel_birth=_, rate_units=_, metadata=None, **kwargs):

        super().__init__()
        self.define_pars(
            birth_rate = ss.peryear(20),
            rel_birth = 1,
            rate_units = 1e-3,  # assumes birth rates are per 1000. If using percentages, switch this to 1
        )
        self.update_pars(pars, **kwargs)

        # Process metadata. Defaults here are the labels used by UN data
        self.metadata = sc.mergedicts(
            sc.objdict(data_cols=dict(year='Year', value='CBR')),
            metadata,
        )

        # Process data, which may be provided as a number, dict, dataframe, or series
        # If it's a number it's left as-is; otherwise it's converted to a dataframe
        self.pars.birth_rate = self.standardize_birth_data()
        self.n_births_this_step = 0 # For results tracking
        self.dist = ss.bernoulli(p=0) # Used to generate random numbers; set to use the birth rate above
        return

    def init_pre(self, sim):
        """ Initialize with sim information """
        super().init_pre(sim)
        if isinstance(self.pars.birth_rate, pd.DataFrame):
            br_year = self.pars.birth_rate[self.metadata.data_cols['year']]
            br_val = self.pars.birth_rate[self.metadata.data_cols['cbr']]
            all_birth_rates = np.interp(self.timevec, br_year, br_val) # This assumes a year timestep -- probably ok?
            self.pars.birth_rate = all_birth_rates
        return

    def standardize_birth_data(self):
        """ Standardize/validate birth rates - handled in an external file due to shared functionality """
        birth_rate = ss.standardize_data(data=self.pars.birth_rate, metadata=self.metadata)
        if isinstance(birth_rate, (pd.Series, pd.DataFrame)):
            return birth_rate.xs(0,level='age')
        if sc.isnumber(birth_rate):
            # If the user has provided a bare number, assume it is per year
            msg = f'Birth rate was specified as a number rather than a rate - assuming it is {birth_rate} per year'
            ss.warn(msg)
            return ss.peryear(birth_rate)
        return birth_rate

    def init_results(self):
        super().init_results()
        self.define_results(
            ss.Result('new',        dtype=int,   scale=True,  summarize_by='sum',  label='New births'),
            ss.Result('cumulative', dtype=int,   scale=True,  summarize_by='last', label='Cumulative births'),
            ss.Result('cbr',        dtype=float, scale=False, summarize_by='mean', label='Crude birth rate'),
        )
        return

    def get_births(self):
        """
        Extract the right birth rates to use and translate it into a number of people to add.
        """
        sim = self.sim
        p = self.pars

        if isinstance(p.birth_rate, (pd.Series, pd.DataFrame)):
            available_years = p.birth_rate.index
            year_ind = sc.findnearest(available_years, sim.t.now('year'))
            nearest_year = available_years[year_ind]
            this_birth_rate = p.birth_rate.loc[nearest_year]
            scaled_birth_prob = ss.prob.array_to_prob(this_birth_rate, self.t.dt, v=p.rate_units * p.rel_birth)
        else:
            this_birth_rate = p.birth_rate
            if isinstance(this_birth_rate, ss.Rate):
                this_birth_rate = ss.prob(this_birth_rate.value * self.pars.rate_units * self.pars.rel_birth, this_birth_rate.unit).to_prob(ss.years(1))
            else: # number
                this_birth_rate = this_birth_rate * self.pars.rate_units * self.pars.rel_birth

            scaled_birth_prob = ss.prob.array_to_prob(np.array([this_birth_rate]), self.t.dt)[0]

        scaled_birth_prob = np.clip(scaled_birth_prob, a_min=0, a_max=1)
        self.dist.set(p=scaled_birth_prob) # Update the distribution with the probabilities for this timestep
        birth_uids = self.dist.filter()
        # n_new = np.random.binomial(n=sim.people.alive.count(), p=scaled_birth_prob) # Not CRN safe, see issue #404
        return birth_uids

    def step(self):
        new_uids = self.add_births()
        self.n_births_this_step = len(new_uids)
        return new_uids

    def add_births(self):
        """ Add n_new births to each state in the sim """
        people = self.sim.people
        birth_uids = self.get_births()
        new_uids = people.grow(len(birth_uids))
        people.age[new_uids] = 0
        people.parent[new_uids] = birth_uids
        return new_uids

    def update_results(self):
        """ Calculate new births and crude birth rate """
        # New births -- already calculated
        self.results.new[self.ti] = self.n_births_this_step

        # Calculate crude birth rate (CBR)
        inv_rate_units = 1.0/self.pars.rate_units
        births_per_year = self.n_births_this_step/self.sim.t.dt_year
        denom = self.sim.people.alive.sum()
        self.results.cbr[self.ti] = inv_rate_units*births_per_year/denom
        return

    def finalize(self):
        super().finalize()
        self.results.cumulative[:] = np.cumsum(self.results.new)
        return


class Deaths(Demographics):
    """
    Configure disease-independent "background" deaths.

    The probability of death for each agent on each timestep is determined
    by the `death_rate` parameter and the time step. The default value of
    this parameter is 0.01, indicating that all agents will
    face a 1% chance of death per year.

    However, this function can be made more realistic by using a dataframe
    for the `death_rate` parameter, to allow it to vary by year, sex, and
    age.  The separate 'metadata' argument can be used to configure the
    details of the input datafile.

    Alternatively, it is possible to override the `death_rate` parameter
    with a bernoulli distribution containing a constant value of function of
    your own design.

    Args:
        pars: dict with arguments including:
            rel_death: constant used to scale all death rates
            death_rate: float, dict, or pandas dataframe/series containing mortality data
            rate_units: units for death rates (see in-line comment on par dict below)

        metadata: data about the data contained within the data input.
            "data_cols" is is a dictionary mapping standard keys, like "year" to the
            corresponding column name in data. Similar for "sex_keys".
    """
    def __init__(self, pars=None, rel_death=_, death_rate=_, rate_units=_, metadata=None, **kwargs):
        super().__init__()
        self.define_pars(
            rel_death = 1,
            death_rate = ss.peryear(10),  # Default = a fixed rate of 2%/year, overwritten if data provided
            rate_units = 1e-3,  # assumes death rates are per 1000. If using percentages, switch this to 1
        )
        self.update_pars(pars, **kwargs)

        # Process metadata. Defaults here are the labels used by UN data
        self.metadata = sc.mergedicts(
            sc.objdict(
                data_cols = dict(year='Time', sex='Sex', age='AgeGrpStart', value='mx'),
                sex_keys = {'Female':'f', 'Male':'m'},
            ),
            metadata
        )

        # Process data, which may be provided as a number, dict, dataframe, or series
        # If it's a number it's left as-is; otherwise it's converted to a dataframe
        self.death_rate_data = self.standardize_death_data() # TODO: refactor
        self._p_death = ss.bernoulli(p=0)  # Placeholder, populated by make_p_death
        self.n_deaths = 0 # For results tracking
        return

    def standardize_death_data(self):
        """ Standardize/validate death rates - handled in an external file due to shared functionality """
        death_rate = ss.standardize_data(data=self.pars.death_rate, metadata=self.metadata)
        if isinstance(death_rate, (pd.Series, pd.DataFrame)):
            death_rate = death_rate.unstack(level='age')
            assert not death_rate.isna().any(axis=None) # For efficiency, we assume that the age bins are the same for all years in the input dataset
        if sc.isnumber(death_rate):
            # If the user has provided a bare number, assume it is per year
            msg = f'Death rate was specified as a number rather than a rate - assuming it is {death_rate} per year'
            ss.warn(msg)
            return ss.peryear(death_rate)
        return death_rate

    def make_p_death(self):
        """ Take in the module, sim, and uids, and return the probability of death for each UID on this timestep """
        sim = self.sim
        drd = self.death_rate_data

        # We are going to set up death_rate as a prob with unit=ss.years(1). That is the probability of death in 1 year
        if sc.isnumber(drd):
            death_rate = np.array([drd * self.pars.rate_units * self.pars.rel_death])  # Later gets interpreted as a prob with unit=ss.years(1) by ss.prob.array_to_prob
        elif isinstance(drd, ss.Rate):
            if drd.unit == 1: # drd is already in years so don't need to translate
                death_rate = np.array([drd.value * self.pars.rate_units * self.pars.rel_death])
            else:
                # Convert from prob per drd.unit to prob per year, ss.prob.array_to_prob will convert to prob per timestep later
                death_rate = np.array([ss.prob(drd.value * self.pars.rate_units * self.pars.rel_death, drd.unit).to_prob(ss.years(1))])

        # Process data
        else:
            ppl = sim.people
            uids = ppl.auids  # Get the UIDs of all alive people

            available_years = drd.index.get_level_values('year')
            year_ind = sc.findnearest(available_years, sim.t.now('year'))  # TODO: make work with different timesteps
            nearest_year = available_years[year_ind]

            death_rate = np.empty(uids.shape, dtype=ss_float)

            if 'sex' in drd.index.names:
                s = drd.loc[nearest_year, 'f']
                binned_ages = np.digitize(ppl.age[ppl.female], s.index)-1 # Negative ages will be in the first bin - do *not* subtract 1 so that this bin is 0
                death_rate[ppl.female] = s.values[binned_ages]
                s = drd.loc[nearest_year, 'm']
                binned_ages = np.digitize(ppl.age[ppl.male], s.index)-1 # Negative ages will be in the first bin - do *not* subtract 1 so that this bin is 0
                death_rate[ppl.male] = s.values[binned_ages]
            else:
                s = drd.loc[nearest_year]
                binned_ages = np.digitize(ppl.age, s.index)-1 # Negative ages will be in the first bin - do *not* subtract 1 so that this bin is 0
                death_rate[:] = s.values[binned_ages]
            death_rate *= self.pars.rate_units * self.pars.rel_death

        # Scale from rate to probability
        death_rate = ss.peryear(death_rate)
        p_death = death_rate.to_prob(self.t.dt)  # Convert to probability per timestep

        if sc.isnumber(drd) or isinstance(drd, ss.Rate):
            p_death = p_death[0] # TODO: what???
        return p_death

    def init_results(self):
        super().init_results()
        self.define_results(
            ss.Result('new',        dtype=int,   scale=True,  summarize_by='sum',  label='Deaths', auto_plot=False), # Use sim deaths instead
            ss.Result('cumulative', dtype=int,   scale=True,  summarize_by='last', label='Cumulative deaths', auto_plot=False),
            ss.Result('cmr',        dtype=float, scale=False, summarize_by='mean', label='Crude mortality rate'),
        )
        return

    def step(self):
        """ Select people to die """
        p_death = self.make_p_death()  # Get the probability of death for each agent
        self._p_death.set(p=p_death)  # Update the distribution with the probabilities for this timestep
        death_uids = self._p_death.filter()
        self.sim.people.request_death(death_uids)
        self.n_deaths = len(death_uids)
        return self.n_deaths

    def update_results(self):
        self.results['new'][self.ti] = self.n_deaths
        return

    def finalize(self):
        super().finalize()
        self.results.cumulative[:] = np.cumsum(self.results.new)
        units = self.pars.rate_units*self.sim.t.dt_year
        inds = self.match_time_inds()
        n_alive = self.sim.results.n_alive[inds]
        deaths = sc.safedivide(self.results.new, n_alive, default=0)
        self.results.cmr[:] = deaths/units
        return


class PregnancyPars(ss.Pars):
    """
    Pregnancy parameters and default values
    """
    def __init__(self, **kwargs):
        super().__init__()

        # Parameters related to probability of getting pregnant
        self.fertility_rate = ss.peryear(100)  # Can be a number or a Pandas DataFrame
        self.rel_fertility = 1  # Constant to scale all fertility rates, useful if a dataframe is used
        self.p_infertile = ss.bernoulli(p=0)  # Primary infertility
        self.min_age = 15  # Minimum age to become pregnant
        self.max_age = 50  # Maximum age to become pregnant
        self.rate_units = 1e-3  # Assumes fertility rates are per 1000. If using percentages, switch this to 1

        # Parameters related to pregnancy duration
        self.dur_pregnancy = ss.choice(a=ss.weeks(np.arange(32, 43)), p=np.array([0.001, 0.002, 0.005, 0.012, 0.026, 0.05, 0.087, 0.134, 0.188, 0.226, 0.269])) # Quantiles for looking up fertility rates at delivery time

        # Parameters related to breastfeeding
        self.dur_breastfeed = ss.lognorm_ex(mean=ss.years(0.75), std=ss.years(0.5))
        self.p_breastfeed = ss.bernoulli(p=1)

        # Parameters related to implantation rate
        self.embryos_per_pregnancy = ss.choice(a=np.array([1, 2]), p=np.array([1.0, 0.0]))  # Embryos per pregnancy

        # Pregnancy outcome parameters - PTBs, deaths, losses
        self.rr_ptb = ss.normal(loc=1, scale=0.1)  # Base risk of pre-term birth due to factors other than maternal age
        self.rr_ptb_age = np.array([[18, 35, 1000], [1.2, 1, 1.2]]) # Relative risk of pre-term birth by maternal age
        self.p_maternal_death = ss.bernoulli(0)
        self.p_survive_maternal_death = ss.bernoulli(0)
        self.p_loss = ss.bernoulli(p=0)     # Per-timestep probability of spontaneous pregnancy loss (miscarriage or stillbirth)
        self.loss_threshold = ss.weeks(20)  # GA cutoff: <20w = miscarriage, >=20w = stillbirth

        # Parameters related to newborn agents
        self.sex_ratio = ss.bernoulli(0.5)  # Ratio of babies born female
        self.slot_scale = 5 # Random slots will be assigned to newborn agents between min=n_agents and max=slot_scale*n_agents
        self.min_slots = 100  # Minimum number of slots, useful if the population size is very small

        # Classification thresholds
        self.preterm_threshold = ss.weeks(37)
        self.very_preterm_threshold = ss.weeks(32)

        # Settings
        self.burnin = True # Should we seed pregnancies that would have happened before the start of the simulation?
        self.trimesters = [ss.weeks(13), ss.weeks(26)]
        self.update(kwargs)
        return


class Pregnancy(Demographics):
    """
    Create births via pregnancies for each agent.

    This module models conception, pregnancy, and birth at the individual level using
    age-specific fertility rates. Supports prenatal and postnatal transmission
    networks, maternal and neonatal mortality, and burn-in of existing
    pregnancies at simulation start.

    Args:
        fertility_rate (float/dataframe): value or dataframe with age-specific fertility rates
        rel_fertility (float): constant used to scale all fertility rates
        p_infertile (bernoulli): probability of primary infertility (default 0)
        min_age (float): minimum age for pregnancy (default 15)
        max_age (float): maximum age for pregnancy (default 50)
        rate_units (float): units for fertility rates (default assumes per 1000)
        dur_pregnancy (float/dur): duration of pregnancy (default: drawn from choice distribution, 32-42 weeks)
        dur_breastfeed (float/dur): duration of breastfeeding (default: drawn from lognormal distribution, 9±6 months)

        rr_ptb (float): base relative risk of pre-term birth (default normal with mean 1, std 0.1)
        rr_ptb_age (array): relative risk of pre-term birth by maternal age (default [[18,35,100],[1.2,1,1.2]])
        p_maternal_death (float): probability of maternal death during pregnancy (default 0.0)
        p_survive_maternal_death (float): probability that an unborn agent will survive death of the mother (default 0)
        sex_ratio (float): probability of female births (default 0.5)
        burnin (bool): whether to seed pregnancies from before simulation start (default true)
        slot_scale (float): scale factor for assigning slots to newborn agents (default 5)
        min_slots (int): minimum number of slots for newborn agents (default 100)
        preterm_threshold (dur): gestational age below which preterm is declared (default 37w)
        very_preterm_threshold (dur): gestational age below which very preterm is declared (default 32w)
        trimesters (list): list of durations defining the end of each trimester
        metadata (dict): data column mappings for fertility rate data if a dataframe is supplied
    """
    def __init__(self, pars=None, fertility_rate=_, rel_fertility=_, p_infertile=_, min_age=_, max_age=_,
                 rate_units=_, dur_pregnancy=_, dur_breastfeed=_, p_breastfeed=_, rr_ptb=_,
                 rr_ptb_age=_, p_maternal_death=_, p_survive_maternal_death=_, sex_ratio=_, burnin=_, slot_scale=_,
                 min_slots=_, trimesters=_, metadata=None, **kwargs):
        super().__init__()
        default_pars = PregnancyPars()
        self.define_pars(**default_pars)
        self.update_pars(pars, **kwargs)

        # Distributions: binary outcomes
        self._p_conceive = ss.bernoulli(p=0)  # Placeholder, see make_p_conceive

        # States
        self.define_states(
            # Pregnancy and fertility
            ss.BoolState('pregnant', label='Pregnant'),  # Currently pregnant
            ss.BoolState('infertile', label='Infertile (primary infertility)', default=self.pars.p_infertile),
            ss.FloatArr('rel_sus', default=0),  # Susceptibility to pregnancy, set to 1 for non-pregnant women
            ss.FloatArr('dur_pregnancy', label='Pregnancy duration'),  # Duration of pregnancy
            ss.FloatArr('parity', label='Parity', default=0),  # Number of births (may include live + still)
            ss.FloatArr('n_pregnancies', label='Number of pregnancies', default=0),  # Number of pregnancies
            ss.FloatArr('n_miscarriages', label='Number of miscarriages', default=0),  # Number of miscarriages per mother
            ss.FloatArr('n_stillbirths', label='Number of stillbirths', default=0),  # Number of stillbirths per mother
            ss.FloatArr('gestation', label='Number of weeks into pregnancy'),  # Gestational clock
            ss.FloatArr('gestation_at_birth', label='Gestational age in weeks'),  # Gestational age at birth, NA if not born during the sim
            ss.FloatArr('ti_pregnant', label='Time of pregnancy'),  # Time pregnancy begins
            ss.FloatArr('ti_delivery', label='Timestep of delivery'),  # Timestep of delivery (integer)
            ss.FloatArr('ti_dead', label='Time of maternal death'),  # Maternal mortality
            ss.BoolState('carrying_multiple', label='Carrying multiple embryos'),  # Carrying multiple embryos

            # Pre-term birth
            ss.FloatArr('rel_ptb_base', label='Relative risk of pre-term birth at base', default=1),  # Baseline relative risk of pre-term birth - constant throughout lifetime of mother
            ss.FloatArr('rel_ptb', label='Relative risk of pre-term birth', default=1),  # Relative risk of pre-term birth - resets every timestep

            # Breastfeeding
            ss.BoolState('breastfeeding', label='Breastfeeding'),  # Currently breastfeeding
            ss.FloatArr('dur_breastfeed', label='Duration of breastfeeding'),  # Duration of breastfeeding
            ss.FloatArr('ti_stop_breastfeed', label='Time breastfeeding stops'),  # Time breastfeeding stops
            ss.BoolState('breastfed', label='Breastfed'),  # Property of newborn indicating whether they were breastfed

            # Birth outcome classification (stored on newborns)
            ss.BoolArr('preterm', label='Preterm birth'),          # <37w GA
            ss.BoolArr('very_preterm', label='Very preterm birth'),  # <32w GA
            ss.BoolArr('neonatal_death', label='Neonatal death'),  # Death within 28 days of birth
        )

        # Process metadata. Defaults here are the labels used by UN data
        self.metadata = sc.mergedicts(
            sc.objdict(data_cols=dict(year='Time', age='AgeGrp', value='ASFR')),
            metadata,
        )
        self.choose_slots = None  # Distribution for choosing slots; set in self.init()
        self.fertility_rate_data = None  # Processed data; set in self.init_pre() if fertility rate data is in pars

        self.derived_results = None  # Updated in init_results

        # Step-level accumulators for events in step() (recorded in update_results, then reset)
        self._counts = sc.objdict(births=0, pregnancies=0, n_preterm=0, n_very_preterm=0)

        # Callbacks for external modules (e.g. FetalHealth)
        self._conception_callbacks = []
        self._delivery_callbacks = []

        # Define ASFR
        self.asfr_bins = np.array([0, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 100])
        self.asfr_width = self.asfr_bins[1]-self.asfr_bins[0]
        self.asfr = None  # Storing this separately from results as it has a different format

        return

    @property
    def fecund(self):
        """ Defined as women of childbearing age """
        ppl = self.sim.people
        pars = self.pars
        return (ppl.age >= pars.min_age) & (ppl.age <= pars.max_age) & ppl.female

    @property
    def fertile(self):
        """ Defined as women of childbearing age who are not infertile. Includes women who may be pregnant """
        return self.fecund & (~self.infertile)

    @property
    def susceptible(self):
        """ Defined as fertile women of childbearing age who are not pregnant, and so are susceptible to conception """
        return self.fertile & (~self.pregnant)

    @property
    def nulliparous(self):
        """ Women of childbearing age who have never been pregnant """
        return (self.parity == 0) & self.fecund

    @property
    def dur_gestation(self):
        """ Return duration of gestation for currently-pregnant women in years """
        return ss.years((self.ti - self.ti_pregnant[self.pregnant])*self.t.dt_year)

    @property
    def dur_gestation_at_birth(self):
        """ Return duration of gestation at birth for agents born during the simulation """
        born_during_sim = self.sim.people.parent.notnan & (self.sim.people.age > self.t.dt.years)
        return ss.weeks(self.people.pregnancy.gestation_at_birth[born_during_sim]).to('years')

    @property
    def tri1_uids(self):
        """ Return UIDs of those in their first trimester """
        return self.pregnant.uids[self.dur_gestation < self.pars.trimesters[0]]

    @property
    def tri2_uids(self):
        """ Return UIDs of those in their second trimester """
        return self.pregnant.uids[(self.dur_gestation >= self.pars.trimesters[0]) & (self.dur_gestation < self.pars.trimesters[1])]

    @property
    def tri3_uids(self):
        """ Return UIDs of those in their third trimester """
        return self.pregnant.uids[self.dur_gestation >= self.pars.trimesters[1]]

    def find_unborn_children(self, parent_uids):
        all_children = self.sim.people.find_children(parent_uids)
        unborn_children = all_children[self.sim.people.age[all_children] < self.t.dt.years]
        return unborn_children

    def make_p_conceive(self, filter_uids=None):
        """ Take in the module, sim, and uids, and return the conception probability for each UID on this timestep """
        ppl = self.sim.people

        # Apply filter UIDS and get ages
        uids = self.fecund
        if filter_uids is not None: uids = filter_uids & uids
        age = ppl.age[uids]

        # Get data, check it's in the right form
        frd = self.fertility_rate_data
        if sc.isnumber(frd):
            raise TypeError('Fertility rate should be specified as a rate (or with time-varying data)')

        # Initialize rate as an array the length of raw UIDs, so we can index it with UIDs
        fertility_rate = np.zeros(len(ppl.uid.raw), dtype=ss_float)

        if isinstance(frd, ss.Rate):
            fertility_rate[uids] = ss.prob(frd.value * (self.pars.rate_units * self.pars.rel_fertility), frd.unit).to_prob(ss.years(1))
        else:
            # Get the time of birth for pregnancies conceived now and pull out the birth rates for that year.
            # We adjust the birth rates to account for the fact that some women are already pregnant, which
            # is why we make a copy.
            birth_year = self.t.year + self.pars.dur_pregnancy.pars.a.years
            birth_year_inds = sc.findnearest(frd.index, birth_year)
            nearest_year = frd.index[birth_year_inds][0]  # 0 because the corner case of spanning 2 years can be ignored
            new_rate = self.fertility_rate_data.loc[nearest_year].values.copy()  # Initialize array with new rates

            # Assign agents to age bins
            age_bins = self.fertility_rate_data.columns.values
            age_bin_all = np.digitize(age, age_bins) - 1

            if self.pregnant.any():
                # Scale the new rate to convert the denominator from all women to non-pregnant women
                v, c = np.unique(age_bin_all, return_counts=True)
                age_counts = np.zeros_like(new_rate)
                age_counts[v] = c

                age_bin_pregnant = np.digitize(ppl.age[self.pregnant], age_bins) - 1
                v, c = np.unique(age_bin_pregnant, return_counts=True)
                pregnant_age_counts = np.zeros_like(new_rate)
                pregnant_age_counts[v] = c

                num_to_make = new_rate * age_counts  # Number that we need to make pregnant
                new_denom = age_counts - pregnant_age_counts  # New denominator for rates
                np.divide(num_to_make, new_denom, where=new_denom>0, out=new_rate)

            # Overall fertility rate
            fertility_rate[uids] = new_rate[age_bin_all] * (self.pars.rate_units * self.pars.rel_fertility)  # Prob per year

        # Scale from rate to probability
        fertility_rate[self.pregnant.uids] = 0  # Currently pregnant women cannot become pregnant
        fertility_rate = ss.peryear(fertility_rate[filter_uids])  # Only return rates for requested UIDs
        p_conceive = fertility_rate.to_prob(self.t.dt)  # Convert to probability per timestep
        return p_conceive

    def standardize_fertility_data(self):
        """
        Standardize/validate fertility rates
        """
        fertility_rate = ss.standardize_data(data=self.pars.fertility_rate, metadata=self.metadata)
        if isinstance(fertility_rate, (pd.Series, pd.DataFrame)):
            fertility_rate = fertility_rate.unstack()
            # Interpolate to 1 year increments
            fertility_rate = fertility_rate.reindex(np.arange(fertility_rate.index.min(), fertility_rate.index.max() + 1)).interpolate()
            max_age = fertility_rate.columns.max()
            fertility_rate[max_age + 1] = 0
            assert not fertility_rate.isna().any(axis=None) # For efficiency, we assume that the age bins are the same for all years in the input dataset
        if sc.isnumber(fertility_rate):
            msg = f'Fertility rate was specified as a number rather than a rate - assuming it is {fertility_rate} per year'
            ss.warn(msg)
            return ss.peryear(fertility_rate)
        return fertility_rate

    def init_pre(self, sim):
        super().init_pre(sim)

        # Process data, which may be provided as a number, dict, dataframe, or series
        # If it's a number it's left as-is; otherwise it's converted to a dataframe
        if self.pars.fertility_rate is not None:
            self.fertility_rate_data = self.standardize_fertility_data()

        low = sim.pars.n_agents + 1
        high = int(self.pars.slot_scale*sim.pars.n_agents)
        high = np.maximum(high, self.pars.min_slots) # Make sure there are at least min_slots slots to avoid artifacts related to small populations
        self.choose_slots = ss.randint(low=low, high=high, sim=sim, module=self)

        return

    def init_post(self):
        super().init_post()
        self.updates_pre()

        # Burn-in
        if self.ti == 0 and self.pars.burnin:  # TODO: refactor
            dist = self.pars.dur_pregnancy
            max_time = dist.rvs(1000).max() # Calculate the maximum duration of pregnancy
            dtis = np.arange(np.ceil(-max_time), 0, 1).astype(int) # e.g. -9, -8 ... -1
            for dti in dtis:
                self.t.ti = dti
                self.step()
            self.t.ti = 0

        return

    def init_results(self):
        """
        Initialize results. By default, this includes:
        - pregnancies: number of new pregnancies on each timestep
        - births: number of new births on each timestep
        - cbr: crude birth rate on each timestep
        - tfr: total fertility rate on each timestep
        - The number of people who are fecund, (in)fertile, susceptible, postpartum, pregnant, infertile, breastfeeding
        """
        super().init_results()

        scaling_kw = dict(dtype=int, scale=True)
        nonscaling_kw = dict(dtype=float, scale=False)
        self.derived_results = ['n_fecund', 'n_fertile', 'n_susceptible']

        # Define results
        results = sc.autolist()
        results += [
            ss.Result('pregnancies',     **scaling_kw, label='New pregnancies', summarize_by='sum'),
            ss.Result('births',          **scaling_kw, label='New births', summarize_by='sum'),
            ss.Result('n_preterm',       **scaling_kw, label='Preterm births', auto_plot=False),
            ss.Result('n_very_preterm',  **scaling_kw, label='Very preterm births', auto_plot=False),
            ss.Result('miscarriages',    **scaling_kw, label='Miscarriages', summarize_by='sum', auto_plot=False),
            ss.Result('stillbirths',     **scaling_kw, label='Stillbirths', summarize_by='sum', auto_plot=False),
            ss.Result('neonatal_deaths', **scaling_kw, label='Neonatal deaths', summarize_by='sum', auto_plot=False),
            ss.Result('maternal_deaths', **scaling_kw, label='Maternal deaths', summarize_by='sum', auto_plot=False),
            ss.Result('preterm_rate',    **nonscaling_kw, label='Preterm birth rate', auto_plot=False),
            ss.Result('mmr',             **nonscaling_kw, summarize_by='mean', label='Maternal mortality rate', auto_plot=False),
            ss.Result('cbr',             **nonscaling_kw, summarize_by='mean', label='Crude birth rate'),
            ss.Result('tfr',             **nonscaling_kw, summarize_by='sum',  label='Total fertility rate'),
        ]
        for res in self.derived_results:
            results += ss.Result(res, **scaling_kw)
        self.define_results(*results)

        # Extra results
        self.asfr = np.zeros((len(self.asfr_bins)-1, self.t.npts))

        return

    # Helper methods
    def _get_uids(self, upper_age=None, female_only=True):
        """ Get the UIDs of people, usually only females, who are below a certain age """
        people = self.sim.people
        if upper_age is None: upper_age = 1000
        within_age = people.age <= upper_age
        if female_only:
            f_uids = (within_age & people.female).uids
            return f_uids
        else:
            uids = within_age.uids
            return uids

    # Methods for setting states
    def set_ptb(self):
        """ Update relative risk of pre-term birth """
        self.rel_ptb[:] = self.rel_ptb_base[:]  # Reset to baseline
        rr_ptb_age_bins = self.pars.rr_ptb_age[0].astype(int)
        age_ind = np.searchsorted(rr_ptb_age_bins, self.sim.people.age, side="left")
        self.rel_ptb[:] *= self.pars.rr_ptb_age[1][age_ind]
        return

    def set_rel_sus(self):
        """
        Set relative susceptibility to pregnancy. Note that rel_sus isn't used in this module,
        but it's a key ingredient for derived modules that compute pregnancies based on exposure.
        """
        self.rel_sus[self.susceptible] = 1.0
        self.rel_sus[~self.susceptible] = 0.0
        return

    def updates_pre(self, uids=None, upper_age=None):
        """
        This runs prior at the beginning of each timestep, prior to calculating pregnancy exposure,
        advancing pregnancies, adding new pregnancies, or determing delivery outcomes. Here we make
        any updates that affect the risk of pregnancy or pre-term birth on this timestep. We also
        set the baseline values for newborn agents.
        """
        if uids is None: uids = self._get_uids(upper_age=upper_age)
        self.infertile[uids] = self.pars.p_infertile.rvs(uids)  # Infertility
        self.rel_ptb_base[uids] = self.pars.rr_ptb.rvs(uids)  # Baseline relative risk of pre-term birth
        self.set_ptb()  # Update pre-term birth relative risk

        # Check if anyone stops breastfeeding
        if self.breastfeeding.any():
            self.update_breastfeeding(self.breastfeeding.uids)

        return uids

    def update_breastfeeding(self, uids):
        stopping = uids[self.ti >= self.ti_stop_breastfeed[uids]]
        if stopping.any():
            self.breastfeeding[stopping] = False
        return stopping

    def process_delivery(self, mother_uids, newborn_uids):
        """
        Handle delivery by updating all states for the mother. This method also transfers
        recorded information stored with the mother to the newborn. During pregnancy,
        gestational age is tracked with the mother; at birth, this value is transferred
        to the newborn agent before being reset for the mother. Likewise, during pregnancy,
        the child UID is stored with the mother; at birth, this value is removed from the mother,
        although newborn agents can still be linked to the mother via the parent state.
        """
        newborn_parents = self.sim.people.parent[newborn_uids]  # Find which parent each newborn belongs to
        self.gestation_at_birth[newborn_uids] = self.gestation[newborn_parents]  # Transfer to newborn

        # Classify preterm births (stored on newborns)
        ga_wk = self.gestation_at_birth[newborn_uids]
        preterm      = ga_wk < self.pars.preterm_threshold.weeks
        very_preterm = ga_wk < self.pars.very_preterm_threshold.weeks
        self.preterm[newborn_uids]      = preterm
        self.very_preterm[newborn_uids] = very_preterm
        self._counts.n_preterm      += preterm.sum()
        self._counts.n_very_preterm += very_preterm.sum()

        self.pregnant[mother_uids] = False
        self.ti_delivery[mother_uids] = self.ti  # Record timestep of delivery as timestep, not fractional time
        self.carrying_multiple[mother_uids] = False  # Reset multiple embryo status
        self.gestation[mother_uids] = np.nan  # No longer pregnant so remove gestational clock
        self.parity[mother_uids] += 1  # Increment parity for the mothers

        # Set breastfeeding status
        self.set_breastfeeding(mother_uids, newborn_uids)  # Update transmission networks

        # Set maternal death outcomes
        dead = self.pars.p_maternal_death.filter(mother_uids)
        self.ti_dead[dead] = self.ti

        # Update networks after the states have been updated above so that the
        # networks can access the latest/current states

        for net in self.sim.networks.values():

            # Update prenatal networks - remove newborns
            if isinstance(net, ss.PrenatalNet):
                net.remove_uids(newborn_uids)

            # Update postnatal networks - add newborns
            if isinstance(net, ss.PostnatalNet):
                net.add_pairs(newborn_parents, newborn_uids)

        # Hook for subclasses to add post-delivery logic
        self._post_delivery(mother_uids, newborn_uids)

        return

    def add_conception_callback(self, fn):
        """ Register a function ``fn(uids)`` to be called after new pregnancies are created """
        self._conception_callbacks.append(fn)
        return

    def add_delivery_callback(self, fn):
        """ Register a function ``fn(mother_uids, newborn_uids)`` to be called after delivery """
        self._delivery_callbacks.append(fn)
        return

    def _post_delivery(self, mother_uids, newborn_uids):
        """Hook for subclasses to add post-delivery logic. Base implementation does nothing."""
        for cb in self._delivery_callbacks:
            cb(mother_uids, newborn_uids)
        return

    def set_breastfeeding(self, mother_uids, newborn_uids):
        """
        Set breastfeeding durations for new mothers. This method could be extended to
        store duration of exclusive breastfeeding, partial breastfeeding, etc, and these
        properties could be stored with the infant for tracking other health outcomes.
        """
        will_breastfeed = self.pars.p_breastfeed.filter(mother_uids)
        self.breastfeeding[will_breastfeed] = True  # For the mother
        self.dur_breastfeed[will_breastfeed] = self.pars.dur_breastfeed.rvs(will_breastfeed)
        self.ti_stop_breastfeed[will_breastfeed] = self.ti + self.dur_breastfeed[will_breastfeed]

        gets_breastfed = np.isin(self.sim.people.parent[newborn_uids], will_breastfeed)
        self.breastfed[newborn_uids[gets_breastfed]] = True
        return

    def process_newborns(self, uids):
        """ Set states for newborn agents """
        return

    def progress_pregnancies(self):
        """
        Update pregnant women: advance gestational clock, apply background pregnancy loss.
        """
        if not self.pregnant.any():
            return

        will_deliver = self.pregnant & (self.ti_delivery <= self.ti)
        self.gestation[self.pregnant] += self.dt.weeks

        # For those delivering, compute actual GA from timing (handles FetalHealth shifts)
        actual_dur = self.ti_delivery[will_deliver] - self.ti_pregnant[will_deliver]
        self.gestation[will_deliver] = (self.t.dt * actual_dur).weeks

        # Background spontaneous pregnancy loss (only for ongoing pregnancies)
        ongoing = self.pregnant & ~will_deliver
        if ongoing.any():
            loss_uids = self.pars.p_loss.filter(ongoing.uids)
            if len(loss_uids):
                unborn = self.find_unborn_children(loss_uids)
                self.sim.people.request_death(unborn)

        return

    def update_maternal_deaths(self):
        """
        Check for maternal deaths
        """
        maternal_deaths = (self.ti_dead <= self.ti).uids
        self.sim.people.request_death(maternal_deaths)
        self.results['maternal_deaths'][self.ti] = len(maternal_deaths)
        return

    def select_conceivers(self, uids=None):
        """ Select people to make pregnant """
        # People eligible to become pregnant. We don't remove pregnant people here, these
        # are instead handled in the fertility_dist logic as the rates need to be adjusted
        if uids is None: uids = self.sim.people.female.uids
        p_conceive = self.make_p_conceive(uids)
        self._p_conceive.set(p_conceive)
        conceive_uids = self._p_conceive.filter(uids)

        if len(conceive_uids) == 0:
            return ss.uids()

        # Validation
        if self.pregnant[conceive_uids].any():
            which_uids = conceive_uids[self.pregnant[conceive_uids]]
            errormsg = f'New conceptions registered in {len(which_uids)} pregnant agent(s) at timestep {self.ti}.'
            raise ValueError(errormsg)

        return conceive_uids

    def _make_newborn_uids(self, conceive_uids, embryo_counts):
        """ Helper method to link embryos to mothers """
        # Choose slots for the first embryo of each pregnancy
        first_slots = self.choose_slots.rvs(conceive_uids)

        # If all pregnancies are singletons, just return
        if np.all(embryo_counts == 1):
            new_uids = self.sim.people.grow(len(first_slots), first_slots)
            return conceive_uids, new_uids, first_slots

        # Otherwise, build the full list of slots for multiple embryos
        new_slots = list(first_slots)

        # Only iterate over pregnancies with multiple embryos to add additional slots
        multi_idx = np.where(embryo_counts > 1)[0]
        insert_offset = 0
        for idx in multi_idx:
            slot = first_slots[idx]
            count = embryo_counts[idx]
            # Insert additional embryos after the first one
            insert_pos = idx + 1 + insert_offset
            for i in range(1, count):
                new_slots.insert(insert_pos, slot + i)
                insert_pos += 1
            insert_offset += count - 1

        # Grow the population to accommodate all embryos
        new_uids = self.sim.people.grow(len(new_slots), new_slots)

        # Expand conceive_uids to match the number of embryos
        conceive_uids_with_repeats = np.repeat(conceive_uids, embryo_counts)

        return conceive_uids_with_repeats, new_uids, new_slots

    def _set_embryo_states(self, conceive_uids, new_uids, new_slots):
        """ Set states for the just-conceived """
        people = self.sim.people
        gest_years = self.dur_pregnancy[conceive_uids] * self.t.dt_year
        people.age[new_uids] = -gest_years
        people.slot[new_uids] = new_slots  # Before sampling female_dist
        people.female[new_uids] = self.pars.sex_ratio.rvs(conceive_uids)
        people.parent[new_uids] = conceive_uids
        return

    def make_embryos(self, conceive_uids, embryo_counts=None):
        """
        Make newly-conceived agents. This method calls two helper methods, which grow the population
        and set the states for the newborn agents.
        """
        people = self.sim.people
        n_unborn = len(conceive_uids)
        if n_unborn == 0:
            new_uids = ss.uids()
        else:
            conceive_uids_with_repeats, new_uids, new_slots = self._make_newborn_uids(conceive_uids, embryo_counts)
            self._set_embryo_states(conceive_uids_with_repeats, new_uids, new_slots)

        if self.ti < 0:
            people.age[new_uids] += -self.ti * self.sim.t.dt_year  # Age to ti=0

        return conceive_uids_with_repeats, new_uids

    def make_pregnancies(self, uids):
        """
        Make pregnancies, assign durations, and determine multiples
        """

        # Change states for the newly pregnant woman
        ti = self.ti
        self.pregnant[uids] = True
        self.ti_pregnant[uids] = ti
        self.gestation[uids] = 0
        self.n_pregnancies[uids] += 1

        # Use rel_ptb to assign pregnancy durations
        rel_ptb = self.rel_ptb[uids]
        sorted_uids = uids[np.argsort(-rel_ptb)]
        dur_preg = self.pars.dur_pregnancy.rvs(uids)
        self.dur_pregnancy[sorted_uids] = np.sort(dur_preg)
        self.ti_delivery[sorted_uids] = ti + self.dur_pregnancy[sorted_uids]

        # Figure out how many embryos each UID will gestate
        # Could do something here to adjust for age, parity, etc.
        # Could also adjust pre-term birth rates based on number of embryos
        embryo_counts = self.pars.embryos_per_pregnancy.rvs(uids)
        self.carrying_multiple[uids[(embryo_counts > 1)]] = True

        # Check that all pregnant women have a delivery time set
        missing_delivery = self.pregnant[uids] & np.isnan(self.ti_delivery[uids])
        if missing_delivery.any():
            which_uids = uids[missing_delivery]
            errormsg = f'Delivery time has NaN values for {len(which_uids)} pregnant agent(s) at timestep {self.ti}.'
            raise ValueError(errormsg)

        return embryo_counts

    def step(self):
        """ Perform all updates except for deaths, which are handled in finish_step """

        # Set base states
        self.updates_pre(upper_age=self.t.dt_year)  # Set base states for new agents

        # Update ongoing pregnancies
        self.progress_pregnancies()

        # Process deliveries and births
        mothers = (self.pregnant & (self.ti_delivery <= self.ti)).uids
        if len(mothers):
            newborns = self.find_unborn_children(mothers)
            self.process_delivery(mothers, newborns)    # Resets maternal states & transfers data to child
            self._counts.births += len(newborns)       # += to handle burn-in
            self.process_newborns(newborns)             # Process newborns

        # Figure out who conceives, set prognoses, and make embryos
        self.set_rel_sus()                              # Update rel_sus
        conceivers = self.select_conceivers()            # Get the UIDs of women who are going to conceive this timestep
        self._counts.pregnancies += len(conceivers)  # += to handle burn-in

        # Make pregnancies and embryos
        if len(conceivers):
            embryo_counts = self.make_pregnancies(conceivers)
            conceive_uids_with_repeats, new_uids = self.make_embryos(conceivers, embryo_counts)

            # Fire conception callbacks
            for cb in self._conception_callbacks:
                cb(conceivers)

            # Add embryos to prenatal networks
            for net in self.sim.networks.values():
                if isinstance(net, ss.PrenatalNet):
                    net.add_pairs(conceive_uids_with_repeats, new_uids)

        # Other updates - maternal deaths
        self.update_maternal_deaths()                   # Handle maternal deaths

        return

    def step_die(self, uids):
        """ Wipe dates and states following death """
        self.pregnant[uids] = False
        self.dur_pregnancy[uids] = np.nan
        self.gestation[uids] = np.nan
        self.ti_delivery[uids] = np.nan
        return

    def finish_step(self):
        super().finish_step()
        death_uids = ss.uids(self.sim.people.ti_dead <= self.ti)
        if len(death_uids) == 0:
            return
        self.process_maternal_deaths(death_uids)
        self.process_prenatal_deaths(death_uids)
        self.process_neonatal_deaths(death_uids)
        return

    def process_maternal_deaths(self, death_uids):
        """
        If any pregnant mothers die, we also process the death of the unborn agent.
        Default probability is set to 1 meaning we assume that unborn children do not survive.
        """
        mother_death_uids = death_uids[self.pregnant[death_uids]]
        if len(mother_death_uids):
            unborn_uids = self.find_unborn_children(mother_death_uids)
            _, unborn_death_uids = self.pars.p_survive_maternal_death.filter(unborn_uids, both=True)
            if len(unborn_death_uids):
                self.sim.people.request_death(unborn_death_uids)
            self.step_die(mother_death_uids)
        return

    def process_prenatal_deaths(self, death_uids):
        """
        If any deaths occur in unborn agents, classify by gestational age:
        - <loss_threshold (default 20w): miscarriage
        - >=loss_threshold: stillbirth
        """
        is_prenatal = self.sim.people.age[death_uids] < 0
        prenatal_death_uids = death_uids[is_prenatal]
        if len(prenatal_death_uids):
            mother_uids = self.sim.people.parent[prenatal_death_uids]
            ga        = self.gestation[mother_uids]
            threshold = self.pars.loss_threshold.weeks

            is_mc = ga < threshold
            self.n_miscarriages[mother_uids[is_mc]] += 1
            self.n_stillbirths[mother_uids[~is_mc]] += 1
            ti = self.ti
            if 0 <= ti < self.t.npts:  # Skip during burn-in (ti < 0) or after final step
                self.results['miscarriages'][ti] += is_mc.sum()
                self.results['stillbirths'][ti]  += (~is_mc).sum()

            singletons = mother_uids[~self.carrying_multiple[mother_uids]]
            self.step_die(singletons)
        return

    def process_neonatal_deaths(self, death_uids):
        """
        Classify deaths of agents aged 0-28 days as neonatal deaths.
        This is passive detection — any cause of death (disease, background mortality,
        etc.) that kills a newborn within 28 days of birth is recorded.
        """
        age_days = ss.years(self.sim.people.age[death_uids]).days
        is_nnd   = (age_days >= 0) & (age_days <= 28)
        nnd_uids = death_uids[is_nnd]
        if len(nnd_uids):
            self.neonatal_death[nnd_uids] = True
            ti = self.ti
            if 0 <= ti < self.t.npts:  # Skip during burn-in (ti < 0) or after final step
                self.results['neonatal_deaths'][ti] += len(nnd_uids)
        return

    def update_results(self):
        super().update_results()
        ti = self.ti
        res = self.results

        # Write step-level counts to results
        for key, val in self._counts.items():
            res[key][ti] = val
        self._counts.update({k: 0 for k in self._counts})

        # Derived rates
        res['preterm_rate'][ti] = sc.safedivide(res['n_preterm'][ti], res['births'][ti])

        for dr in self.derived_results:
            state = getattr(self, dr.replace('n_', ''))
            res[dr][ti] = state.sum()

        # Update ASFR, TFR, and MMR
        self.compute_asfr()
        res.tfr[ti] = sum(self.asfr[:, ti])*self.asfr_width/1000
        res.mmr[ti] = sc.safedivide(res.maternal_deaths[ti], res.births[ti]) * 100e3

        return

    def compute_asfr(self):
        """
        Computes age-specific fertility rates (ASFR). Since this is calculated each timestep,
        the annualized results should compute the sum.
        """
        new_mother_uids = (self.ti_delivery == self.ti).uids
        new_mother_ages = self.sim.people.age[new_mother_uids]
        births_by_age, _ = np.histogram(new_mother_ages, bins=self.asfr_bins)
        women_by_age, _ = np.histogram(self.sim.people.age[self.sim.people.female], bins=self.asfr_bins)
        self.asfr[:, self.ti] = sc.safedivide(births_by_age, women_by_age) * 1000
        return

    def finalize(self):
        super().finalize()
        units = self.pars.rate_units*self.sim.t.dt_year
        inds = self.match_time_inds()
        n_alive = self.sim.results.n_alive[inds]
        births = sc.safedivide(self.results.births, n_alive, default=0)
        self.results.cbr[:] = births/units

        # Aggregate the ASFR results, taking rolling annual sums
        n_bins = len(self.asfr_bins) - 1
        tdim = min(int(1 / self.t.dt_year), self.t.npts) # Handle sims shorter than 1 year
        cs = np.cumsum(self.asfr, axis=1)
        asfr = np.zeros((n_bins, self.t.npts))
        asfr[:, tdim-1:] = cs[:, tdim-1:] - np.hstack([np.zeros((n_bins, 1)), cs[:, :-tdim]])
        self.asfr = asfr
        return


# %% Fetal health

class FetalHealth(ss.Module):
    """
    Track fetal health outcomes during pregnancy.

    Works alongside the Pregnancy module to model birth weight outcomes (low
    birth weight, small for gestational age) based on fetal growth restriction.
    Disease-agnostic by design: external modules (connectors, interventions)
    modify fetal health via the public API methods.

    Preterm classification is handled by the Pregnancy module (based on
    gestational age at birth). This module focuses on the weight/growth axis.

    Integrates with Pregnancy via callbacks: Pregnancy calls ``on_conception``
    when new pregnancies begin and ``on_delivery`` when births occur. External
    modules (e.g. disease connectors) can register their own callbacks via
    ``add_conception_callback`` to act on new pregnancies after baseline
    initialization.

    During pregnancy, ``weight_percentile``, ``growth_restriction``,
    ``timing_shift``, and ``n_exposures`` are tracked on the mother. At
    delivery, ``birth_weight``, ``lbw``, ``vlbw``, and ``sga`` are stored
    on the newborn.

    Each pregnancy gets a baseline weight percentile drawn at conception.
    Two modification levers are available:

        1. **Delivery timing**: bring ``ti_delivery`` forward (preterm birth risk)
        2. **Growth restriction**: accumulate fractional weight reduction

    At delivery: ``birth_weight = baseline_for_GA × percentile × (1 - restriction)``

    Args:
        weight_by_ga (array):           Nx2 array of [gestational_age_weeks, weight_grams]
        interp_fn (callable):           interpolation function with signature (x, xp, fp) -> array (default np.interp)
        sga_ratio (float):              fraction of GA-appropriate weight below which SGA is declared
        lbw_threshold (float):          birth weight in grams below which LBW is declared
        vlbw_threshold (float):         birth weight in grams below which VLBW is declared
        min_ga (dur):                   floor for timing shifts (delivery can't be brought before this GA)
        percentile_dist (Dist):         distribution for baseline fetal weight percentile

    **Example**::

        import starsim as ss

        sim = ss.Sim(
            demographics=[ss.Pregnancy(fertility_rate=10), ss.Deaths(death_rate=10)],
            modules=ss.FetalHealth(),
            networks=ss.PrenatalNet(),
        )
        sim.run()
    """

    # Approximate 50th percentile fetal weight (grams) by gestational age (weeks)
    # Source: Hadlock 1991 / INTERGROWTH-21st
    default_weight_by_ga = np.array([
        [24, 600],  [25, 700],  [26, 800],  [27, 900],  [28, 1000],
        [29, 1150], [30, 1300], [31, 1500], [32, 1700], [33, 1900],
        [34, 2100], [35, 2400], [36, 2600], [37, 2850], [38, 3050],
        [39, 3250], [40, 3400], [41, 3500], [42, 3550],
    ], dtype=float)

    def __init__(self, pars=None, **kwargs):
        super().__init__(name='fetal_health')
        self.define_pars(
            weight_by_ga=self.default_weight_by_ga,
            interp_fn=np.interp,  # Interpolation function for mapping gestational age to reference birth weight
            sga_ratio=0.87,  # SGA if birth_weight < baseline_for_ga * sga_ratio; ~10% baseline rate given percentile_dist=N(1,0.1)
            lbw_threshold=2500,
            vlbw_threshold=1500,
            min_ga=ss.weeks(24),  # Earliest possible delivery GA; timing shifts cannot bring delivery before this
            percentile_dist=ss.normal(loc=1.0, scale=0.1),
        )
        self.update_pars(pars, **kwargs)

        self.define_states(
            # Pregnancy-time tracking (stored on mothers)
            ss.FloatArr('weight_percentile', label='Fetal weight percentile'),
            ss.FloatArr('growth_restriction', label='Cumulative growth restriction'),
            ss.FloatArr('timing_shift',      label='Accumulated delivery shift (weeks)'),
            ss.FloatArr('n_exposures',       label='Disease exposures during pregnancy'),

            # Birth outcomes (stored on newborns at delivery)
            ss.FloatArr('birth_weight', label='Birth weight (grams)'),
            ss.BoolArr('lbw',  label='Low birth weight'),
            ss.BoolArr('vlbw', label='Very low birth weight'),
            ss.BoolArr('sga',  label='Small for gestational age'),
            ss.BoolArr('svn',  label='Small vulnerable newborn'),
        )

        self._conception_callbacks = []
        return

    def init_pre(self, sim):
        """ Register callbacks with the Pregnancy module """
        super().init_pre(sim)
        if not hasattr(sim.demographics, 'pregnancy'):
            raise ValueError('FetalHealth requires a Pregnancy module. Add ss.Pregnancy() to demographics.')
        preg = sim.demographics.pregnancy
        preg.add_conception_callback(self.on_conception)
        preg.add_delivery_callback(self.on_delivery)
        return

    def add_conception_callback(self, fn):
        """
        Register a function to be called when new pregnancies are detected.
        The function receives ``(uids,)`` after baseline initialization.
        """
        self._conception_callbacks.append(fn)
        return

    def init_results(self):
        super().init_results()
        self.define_results(
            ss.Result('n_births',          dtype=int, label='Births'),
            ss.Result('n_lbw',             dtype=int, label='Low birth weight'),
            ss.Result('n_vlbw',            dtype=int, label='Very low birth weight'),
            ss.Result('n_sga',             dtype=int, label='Small for gestational age'),
            ss.Result('n_svn',             dtype=int, label='Small vulnerable newborns'),
            ss.Result('mean_birth_weight', scale=False, label='Mean birth weight (g)'),
            ss.Result('mean_ga_at_birth',  scale=False, label='Mean GA at birth (weeks)'),
            ss.Result('mean_exposures',    scale=False, label='Mean exposures per pregnancy'),
            ss.Result('lbw_rate',          scale=False, label='LBW rate'),
            ss.Result('sga_rate',          scale=False, label='SGA rate'),
            ss.Result('svn_rate',          scale=False, label='SVN rate'),
        )
        return

    def on_conception(self, uids):
        """ Initialize fetal health for new pregnancies (called by Pregnancy) """
        self.weight_percentile[uids] = self.pars.percentile_dist.rvs(uids)
        self.growth_restriction[uids] = 0.0
        self.timing_shift[uids] = 0.0
        self.n_exposures[uids] = 0
        for cb in self._conception_callbacks:
            cb(uids)
        return

    def on_delivery(self, mother_uids, newborn_uids):
        """
        Classify birth outcomes (called by Pregnancy).

        Birth weight is computed from the mother's pregnancy-time states
        (weight_percentile, growth_restriction). Outcomes (birth_weight,
        lbw, vlbw, sga) are stored on the newborn agents.
        """
        if not len(newborn_uids):
            return

        # Compute birth weight using mother states (handles twins via parent lookup)
        parents = self.sim.people.parent[newborn_uids]
        birth_weights, ga_wk = self.compute_birth_weight(parents)

        # Store outcomes on newborns
        self.birth_weight[newborn_uids] = birth_weights

        ref = self.pars.weight_by_ga
        sga_threshold = self.pars.interp_fn(ga_wk, ref[:, 0], ref[:, 1]) * self.pars.sga_ratio
        lbw  = birth_weights < self.pars.lbw_threshold
        vlbw = birth_weights < self.pars.vlbw_threshold
        sga  = birth_weights < sga_threshold

        self.lbw[newborn_uids]  = lbw
        self.vlbw[newborn_uids] = vlbw
        self.sga[newborn_uids]  = sga

        # SVN: small vulnerable newborn = preterm | lbw | sga
        preterm = self.sim.demographics.pregnancy.preterm[newborn_uids]
        svn = preterm | lbw | sga
        self.svn[newborn_uids] = svn

        # Results
        n  = len(newborn_uids)
        ti = self.ti
        self.results['n_births'][ti]          += n
        self.results['n_lbw'][ti]             += lbw.sum()
        self.results['n_vlbw'][ti]            += vlbw.sum()
        self.results['n_sga'][ti]             += sga.sum()
        self.results['n_svn'][ti]             += svn.sum()
        self.results['mean_birth_weight'][ti]  = birth_weights.mean() if n else 0
        self.results['mean_ga_at_birth'][ti]   = ga_wk.mean() if n else 0
        self.results['mean_exposures'][ti]     = self.n_exposures[parents].mean() if n else 0
        self.results['lbw_rate'][ti]           = lbw.mean() if n else 0
        self.results['sga_rate'][ti]           = sga.mean() if n else 0
        self.results['svn_rate'][ti]           = svn.mean() if n else 0
        return

    def apply_timing_shift(self, uids, shift_weeks):
        """
        Bring delivery forward for pregnant women.

        Uses a one-way ratchet: delivery can only be brought forward, never
        pushed back. The actual shift applied is tracked in ``timing_shift``.

        Args:
            uids: UIDs of pregnant women
            shift_weeks (float/array): shift in weeks; positive = earlier delivery
        """
        if not len(uids):
            return

        preg = self.sim.people.pregnancy
        weeks_per_ts = self.dt.weeks
        shifts_ts = shift_weeks / weeks_per_ts

        min_ga_ts = self.pars.min_ga.weeks / weeks_per_ts
        current_delivery = preg.ti_delivery[uids]
        new_delivery = current_delivery - shifts_ts
        min_delivery = preg.ti_pregnant[uids] + min_ga_ts
        new_delivery = np.maximum(new_delivery, min_delivery)

        actually_shifted_ts = np.maximum(0, current_delivery - new_delivery)
        preg.ti_delivery[uids] = np.minimum(current_delivery, new_delivery)
        self.timing_shift[uids] += actually_shifted_ts * weeks_per_ts
        return

    def apply_growth_restriction(self, uids, penalty):
        """
        Apply fractional growth restriction (cumulative, diminishing).

        Positive penalties use diminishing returns: ``current + (1-current) * penalty``.
        Negative penalties (growth boost, e.g. GDM macrosomia) are additive.

        Args:
            uids: UIDs of pregnant women
            penalty (float): fractional weight reduction; negative = growth boost
        """
        if not len(uids):
            return

        current = self.growth_restriction[uids]
        positive = penalty >= 0
        new_val = np.where(positive, current + (1 - current) * penalty, current + penalty)
        self.growth_restriction[uids] = new_val
        return

    def reverse_timing_shift(self, uids, fraction):
        """
        Recover a fraction of the accumulated delivery timing shift.

        Args:
            uids: UIDs of pregnant women
            fraction (float/array): fraction to recover (0-1)
        """
        if not len(uids):
            return

        preg = self.sim.people.pregnancy
        current_shift = self.timing_shift[uids]
        recover_weeks = current_shift * fraction

        has_shift = recover_weeks > 0
        if not has_shift.any():
            return

        recover_uids = uids[has_shift]
        recover_wk   = recover_weeks[has_shift]
        recover_ts   = recover_wk / self.dt.weeks

        current_delivery = preg.ti_delivery[recover_uids]
        preg.ti_delivery[recover_uids] = current_delivery + recover_ts
        self.timing_shift[recover_uids] -= recover_wk
        return

    def reverse_growth_restriction(self, uids, amount):
        """
        Reverse a specific amount of growth restriction.

        Args:
            uids: UIDs of pregnant women
            amount (float/array): amount to reverse
        """
        if not len(uids):
            return

        current = self.growth_restriction[uids]
        self.growth_restriction[uids] = np.maximum(0, current - amount)
        return

    def compute_birth_weight(self, uids):
        """
        Compute birth weight at delivery.

        Override this method to customize the birth weight formula. The
        interpolation function can be swapped via ``pars.interp_fn``.

        Returns:
            tuple: (birth_weights, ga_weeks) arrays
        """
        preg = self.sim.people.pregnancy
        ga_wk = (preg.ti_delivery[uids] - preg.ti_pregnant[uids]) * self.dt.weeks

        ref = self.pars.weight_by_ga
        baseline    = self.pars.interp_fn(ga_wk, ref[:, 0], ref[:, 1])
        percentile  = self.weight_percentile[uids]
        restriction = self.growth_restriction[uids]

        return baseline * percentile * (1 - restriction), ga_wk

    def step(self):
        pass  # All logic is driven by Pregnancy callbacks



"""
Test connectors and custom interventions
"""

import sciris as sc
import numpy as np
import matplotlib.pyplot as plt
import starsim as ss
import starsim_examples as sse

sc.options(interactive=False) # Assume not running interactively


class hiv_syph(ss.Connector):
    """ Simple connector whereby rel_sus to NG doubles if CD4 count is <200"""
    def __init__(self, **kwargs):
        super().__init__()
        self.define_pars(
            label = 'HIV-Syphilis',
            rel_sus_syph_hiv    = 2,   # People with HIV are 2x more likely to acquire syphilis
            rel_sus_syph_aids   = 5,   # People with AIDS are 5x more likely to acquire syphilis
            rel_trans_syph_hiv  = 1.5, # People with HIV are 1.5x more likely to transmit syphilis
            rel_trans_syph_aids = 3,   # People with AIDS are 3x more likely to transmit syphilis
            rel_sus_hiv_syph    = 2.7, # People with syphilis are 2.7x more likely to acquire HIV
            rel_trans_hiv_syph  = 2.7, # People with syphilis are 2.7x more likely to transmit HIV
        )
        self.update_pars(**kwargs)
        return

    def step(self):
        """ Specify HIV-syphilis interactions """

        diseases = self.sim.diseases
        syph = diseases.syphilis
        hiv = diseases.hiv
        cd4 = self.sim.people.hiv.cd4

        # People with HIV are more likely to acquire syphilis
        syph.rel_sus[cd4 < 500] = self.pars.rel_sus_syph_hiv
        syph.rel_sus[cd4 < 200] = self.pars.rel_sus_syph_aids

        # People with HIV are more likely to transmit syphilis
        syph.rel_trans[cd4 < 500] = self.pars.rel_trans_syph_hiv
        syph.rel_trans[cd4 < 200] = self.pars.rel_trans_syph_aids

        # People with syphilis are more likely to acquire HIV
        hiv.rel_sus[syph.active] = self.pars.rel_sus_hiv_syph

        # People with syphilis are more likely to transmit HIV
        hiv.rel_trans[syph.active] = self.pars.rel_trans_hiv_syph
        return


class Penicillin(ss.Intervention):
    """ Create a penicillin (BPG) intervention for treating syphilis """
    def __init__(self, year=2020, prob=0.8):
        super().__init__() # Initialize the intervention
        self.prob = prob # Store the probability of treatment
        self.year = ss.date(year)
        return

    def step(self):
        sim = self.sim
        if sim.now > self.year:
            syphilis = sim.diseases.syphilis

            # Define who is eligible for treatment
            eligible_ids = syphilis.infected.uids  # People are eligible for treatment if they have just started exhibiting symptoms
            n_eligible = len(eligible_ids) # Number of people who are eligible

            # Define who receives treatment
            is_treated = np.random.rand(n_eligible) < self.prob  # Define which of the n_eligible people get treated by comparing np.random.rand() to self.p
            treat_ids = eligible_ids[is_treated]  # Pull out the IDs for the people receiving the treatment
            syphilis.infected[treat_ids] = False
            syphilis.susceptible[treat_ids] = True
            sim.diseases.hiv.rel_sus[treat_ids] = 1
            sim.diseases.hiv.rel_trans[treat_ids] = 1
        return


def make_args():
    """ Make people, HIV, syphilis, and network """
    pars = dict(n_agents=2000, verbose=0)
    mf = ss.MFNet(duration=ss.lognorm_ex(mean=5, std=0.5, unit=ss.years)) # TODO: think about whether these should be ss.dur(); currently they are not since stored in natural units with -self.dt
    hiv = sse.HIV(beta={'mf': [0.0008, 0.0004]}, init_prev=0.2) # TODO: beta should wrap the other way
    syph = sse.Syphilis(beta={'mf': [0.1, 0.05]}, init_prev=0.05)
    args = dict(pars=pars, networks=mf, diseases=[hiv, syph])
    return args


@sc.timer()
def test_connectors(do_plot=False):
    """ Test connector example """
    sc.heading('Testing connectors')

    # Make arguments
    args = make_args()
    sims = sc.objdict() # List of sims

    # Make a sim with a connector, and run
    sims.con = ss.Sim(label='With connector', connectors=hiv_syph(), **args)
    sims.con.run()

    # Make a sim without a connector, and run
    sims.nocon = ss.Sim(label='Without connector', **args)
    sims.nocon.run()

    # Make a sim with a connector and syph treatment, and run
    sims.treat = ss.Sim(label='With treatment', connectors=hiv_syph(), interventions=Penicillin(), **args)
    sims.treat.run()

    # Parse results
    results = sc.odict()
    diseases = ['syphilis', 'hiv']
    for sim in sims.values():
        results[sim.label] = sc.objdict()
        for disease in diseases:
            results[sim.label][disease] = sim.results[disease].n_infected

    # Plot
    if do_plot:
        plt.figure()

        plt.subplot(2,1,1)
        x = sims.con.t.yearvec
        for label,res in results.items():
            plt.plot(x, res.syphilis, label=label)
        plt.title('Syphilis infections')
        plt.xlabel('Year')
        plt.ylabel('Count')
        plt.axvline(2020)
        plt.legend()

        plt.subplot(2,1,2)
        for label,res in results.items():
            plt.plot(x, res.hiv, label=label)
        plt.title('HIV infections')
        plt.xlabel('Year')
        plt.ylabel('Count')
        plt.axvline(2020)
        plt.legend()

        sc.figlayout()
        plt.show()

    # Check results
    for disease in diseases:
        assert results[0][disease].sum() > results[1][disease].sum(), f'{disease.title()} infections should be higher with connector'
        assert results[0][disease].sum() > results[2][disease].sum(), f'{disease.title()} infections should be lower with treatment'

    return sims


if __name__ == '__main__':
    do_plot = True
    sc.options(interactive=do_plot)
    T = sc.timer()

    sims = test_connectors(do_plot=do_plot)

    T.toc()




"""
Define default gonorrhea disease module and related interventions
"""

import numpy as np
import starsim as ss


__all__ = ['Gonorrhea']

class Gonorrhea(ss.Infection):

    def __init__(self, pars=None, **kwargs):
        # Parameters
        super().__init__()
        self.define_pars(
            beta = ss.prob(1.0), # Placeholder value
            dur_inf_in_days = ss.lognorm_ex(mean=ss.days(10), std=ss.days(0.6)),  # median of 10 days (IQR 7–15 days) https://sti.bmj.com/content/96/8/556
            p_symp    = ss.bernoulli(p=0.5),  # Share of infections that are symptomatic. Placeholder value
            p_clear   = ss.bernoulli(p=0.2),  # Share of infections that spontaneously clear: https://sti.bmj.com/content/96/8/556
            init_prev = ss.bernoulli(p=0.1),
        )
        self.update_pars(pars, **kwargs)

        # States additional to the default disease states (see base class)
        # Additional states dependent on parameter values, e.g. self.p_symp?
        # These might be useful for connectors to target, e.g. if HIV reduces p_clear
        self.define_states(
            ss.BoolState('symptomatic', label='Symptomatic'),
            ss.FloatArr('ti_clearance', label='Time of clearance'),
            ss.FloatArr('p_symp', default=1, label='Probability of symptoms'),
        )

        return

    def init_results(self):
        """ Initialize results """
        super().init_results()
        self.define_results(
            ss.Result('new_clearances', dtype=int, label='New clearances')
        )
        return

    def update_results(self):
        super().update_results()
        ti = self.ti
        self.results.n_symptomatic[ti] = self.symptomatic.count()
        self.results.new_clearances[ti] = np.count_nonzero(self.ti_clearance == ti)
        return

    def step_state(self):
        """ Natural clearance """
        clearances = self.ti_clearance <= self.ti
        self.susceptible[clearances] = True
        self.infected[clearances] = False
        self.symptomatic[clearances] = False
        self.ti_clearance[clearances] = self.ti

        return

    def set_prognoses(self, uids, sources=None):
        """ Natural history of gonorrhea for adult infection """
        super().set_prognoses(uids, sources)
        ti = self.ti

        # Set infection status
        self.susceptible[uids] = False
        self.infected[uids] = True
        self.ti_infected[uids] = ti

        # Set infection status
        symp_uids = self.pars.p_symp.filter(uids)
        self.symptomatic[symp_uids] = True

        # Set natural clearance
        clear_uids = self.pars.p_clear.filter(uids)
        dur = ti + self.pars.dur_inf_in_days.rvs(clear_uids) # Convert from days to years and then adjust for dt automatically # TODO: check!
        self.ti_clearance[clear_uids] = dur
        return


"""
Example neonatal sepsis disease.

A simple SIR-like disease that infects newborns at birth and kills a fraction
shortly after, producing neonatal deaths that the Pregnancy module passively
detects and classifies.

This demonstrates:
    - How a disease can cause neonatal deaths via ``request_death`` on newborns
    - How the Pregnancy module passively detects deaths of agents aged 0-28 days
      and classifies them as neonatal deaths (no special hookup needed)
    - How to use ``set_prognoses`` to apply lethal outcomes only to neonates

The key insight is that neonatal death detection is passive: any mechanism that
kills an agent aged 0-28 days (disease, congenital outcomes, background
mortality) is automatically classified as a neonatal death by the Pregnancy
module. No registration or callback is needed.

Usage::

    import starsim as ss
    import starsim_examples as sse

    sim = ss.Sim(
        diseases=sse.NeonatalSepsis(),
        demographics=[ss.Pregnancy(fertility_rate=ss.freqperyear(30)), ss.Deaths()],
        networks=[ss.PrenatalNet(), ss.RandomNet()],
    )
    sim.run()

    # NND results are on the Pregnancy module, not the disease
    print('Neonatal deaths:', sim.results.pregnancy.nnds.sum())
    print('Total births:', sim.results.pregnancy.births.sum())
"""

import starsim as ss

__all__ = ['NeonatalSepsis']


class NeonatalSepsis(ss.SIR):
    """
    Minimal neonatal sepsis model.

    Infects newborns at birth with probability ``init_prev``, then kills a
    fraction (``p_death``) within a short window (``dur_inf``). Useful for
    testing passive neonatal death detection in the Pregnancy module.

    Pars:
        beta (float):       transmission rate — set to 0 since infection is only at birth
        init_prev (Dist):   fraction of newborns infected at birth
        dur_inf (Dist):     time from infection to death (for those who die)
        p_death (Dist):     case fatality rate among infected neonates
    """

    def __init__(self, pars=None, **kwargs):
        super().__init__()
        self.define_pars(
            beta      = 0,                                         # No ongoing transmission — infection is set at birth
            init_prev = ss.bernoulli(p=0.3),                       # 30% of newborns infected at birth
            dur_inf   = ss.lognorm_ex(mean=ss.days(7), std=ss.days(3)),  # ~1 week illness
            p_death   = ss.bernoulli(p=0.5),                       # 50% case fatality
        )
        self.update_pars(pars, **kwargs)
        return

    def set_prognoses(self, uids, sources=None):
        """
        Set prognoses — only apply lethal outcomes to neonates (age < 28 days).

        Non-neonates who get infected (via init_prev at sim start) just recover
        normally via the base SIR logic.
        """
        super().set_prognoses(uids, sources)

        # Only neonates can die from this disease
        age_days = ss.years(self.sim.people.age[uids]).days
        neonates = uids[age_days <= 28]
        if len(neonates):
            will_die = self.pars.p_death.filter(neonates)
            dur = self.pars.dur_inf.rvs(will_die)
            self.ti_dead[will_die] = self.ti + dur
        return



"""
Fuller port of laser_example_orig.py, using Starsim's naming conventions

Requires starsim>=3.0
"""
import sciris as sc
import numpy as np
import matplotlib.pyplot as plt
import starsim as ss


class SIRModel(ss.Module):
    def __init__(self):
        super().__init__()
        self.define_states(
            ss.Arr('disease_state', dtype=np.int32, default=0),
            ss.Arr('recovery_timer', dtype=np.int32, default=0),
        )

    def init_results(self):
        super().init_results()
        self.define_results(
            ss.Result('S'),
            ss.Result('I'),
            ss.Result('R'),
        )

    def step(self):
        ti = self.ti
        susceptible = (self.disease_state == 0).sum()
        infected = (self.disease_state == 1).sum()
        recovered = (self.disease_state == 2).sum()
        total = len(self.sim.people)
        self.results.S[ti] = susceptible / total
        self.results.I[ti] = infected / total
        self.results.R[ti] = recovered / total

    def plot_results(self):
        plt.figure(figsize=(10, 6))
        plt.plot(self.results.S, label="Susceptible (S)", color="blue")
        plt.plot(self.results.I, label="Infected (I)", color="red")
        plt.plot(self.results.R, label="Recovered (R)", color="green")
        plt.title("SIR Model Dynamics with LASER Components")
        plt.xlabel("Time (Timesteps)")
        plt.ylabel("Fraction of Population")
        plt.legend()
        plt.grid()
        plt.show()
        plt.savefig("gpt_sir.png")


class IntrahostProgression(ss.Module):
    def init_post(self):
        super().init_post()
        n_agents = self.sim.pars.n_agents
        sir = self.sim.people.sirmodel

        # Seed the infection
        num_initial_infected = int(0.01 * n_agents)  # e.g., 1% initially infected
        infected_indices = np.random.choice(n_agents, size=num_initial_infected, replace=False)
        sir.disease_state[infected_indices] = 1

        # Initialize recovery timer for initially infected individuals
        initially_infected = sir.disease_state == 1
        sir.recovery_timer[initially_infected] = np.random.randint(5, 15, size=initially_infected.sum())

    def step(self):
        sir = self.sim.people.sirmodel

        infected = sir.disease_state == 1

        # Decrement recovery timer
        sir.recovery_timer[infected] -= 1

        # Recover individuals whose recovery_timer has reached 0
        recoveries = infected & (sir.recovery_timer <= 0)
        sir.disease_state[recoveries] = 2


class Transmission(ss.Module):
    def __init__(self, infection_rate):
        super().__init__()
        self.infection_rate = infection_rate

    def step(self):
        sir = self.sim.people.sirmodel
        susceptible = sir.disease_state == 0
        infected = sir.disease_state == 1

        num_susceptible = susceptible.sum()
        num_infected = infected.sum()
        population_size = len(self.sim.people)

        # Fraction of infected and susceptible individuals
        fraction_infected = num_infected / population_size

        # Transmission logic: Probability of infection per susceptible individual
        infection_probability = self.infection_rate * fraction_infected

        # Apply infection probability to all susceptible individuals
        new_infections = np.random.rand(num_susceptible) < infection_probability

        # Set new infections and initialize their recovery_timer
        susceptible_indices = np.where(susceptible)[0]
        newly_infected_indices = susceptible_indices[new_infections]
        sir.disease_state[newly_infected_indices] = 1
        sir.recovery_timer[newly_infected_indices] = np.random.randint(5, 15, size=newly_infected_indices.size)  # Random recovery time


def run_sim(run=True, plot=False):
    """ Create and run the simulation """
    pars = dict(
        n_agents = 100_000*2,
        dur = 100*10,
        verbose = 0,
    )
    infection_rate = 0.3

    intrahost = IntrahostProgression()
    transmission = Transmission(infection_rate)
    sir_model = SIRModel()

    # Initialize the model
    sim = ss.Sim(pars, modules=[intrahost, transmission, sir_model], copy_inputs=False)

    # Run the simulation
    if run:
        sim.run()

    # Plot results
    if plot:
        sir_model.plot_results()
    return sim


if __name__ == '__main__':
    with sc.timer():
        sir_model = run_sim()
    sir_model.plot_results()


#!/usr/bin/env python
"""
Run all the Jupyter notebook user guides. Usage:
  ./check_user_guides.py

See docutils.py in the parent folder for more information.
"""
import os
import sys
import sciris as sc
parent = sc.path(__file__).resolve().parent.parent # Ensure it's the real path if it's symlinked
ut = sc.importbypath(parent / 'docutils.py')

if __name__ == '__main__':
    results = ut.execute_notebooks(*sys.argv[1:], folders=['user_guide'])



"""
Define cholera model.
Adapted from https://github.com/optimamodel/gavi-outbreaks/blob/main/stisim/gavi/cholera.py
Original version by Dom Delport 2024
"""

import numpy as np
import starsim as ss

__all__ = ['Cholera']


class Cholera(ss.Infection):
    """
    Cholera
    """
    def __init__(self, pars=None, **kwargs):
        """ Initialize with parameters """
        super().__init__()
        self.define_pars(
            # Initial conditions and beta
            beta = ss.prob(1.0, ss.days(1)), # Placeholder value
            init_prev = ss.bernoulli(0.005),

            # Natural history parameters, all specified in days
            dur_exp2inf   = ss.lognorm_ex(mean=ss.days(2.772), std=ss.days(4.737)), # Calculated from Azman et al. estimates https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3677557/
            dur_asymp2rec = ss.uniform(low=ss.days(1), high=ss.days(10)), # From WHO cholera fact sheet, asymptomatic individuals shed bacteria for 1-10 days (https://www.who.int/news-room/fact-sheets/detail/cholera)
            dur_symp2rec  = ss.lognorm_ex(mean=ss.days(5), std=ss.days(1.8)), # According to Fung most modelling studies assume 5 days of symptoms (https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3926264/), but found a range of 2.9-14 days. Distribution approximately fit to these values
            dur_symp2dead = ss.lognorm_ex(mean=ss.days(1), std=ss.days(0.5)), # There does not appear to be differences in timing/duration of mild vs severe disease, but death from severe disease happens rapidly https://www.ncbi.nlm.nih.gov/pmc/articles/PMC5767916/
            p_death       = ss.bernoulli(p=0.005), # Probability of death is typically less than 1% when treated
            p_symp        = ss.bernoulli(p=0.5), # Proportion of infected which are symptomatic, mid range of ~25% and 57% estimates from Jaclson et al (https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3795095/) and Nelson et al (https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3842031/), respectively
            asymp_trans   = 0.01, # Reduction in transmission probability for asymptomatic infection, asymptomatic carriers shed 100-1000 times less bacteria than symptomatic carriers (https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3084143/ and https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3842031/). Previous models assume a 10% relative transmissibility (https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4238032/)

            # Environmental parameters
            beta_env = ss.prob(0.5 / 3), # Scaling factor for transmission from environment,
            half_sat_rate = 1_000_000, # Infectious dose in water sufficient to produce infection in 50% of exposed, from Mukandavire et al. (https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3102413/)
            shedding_rate = ss.freqperday(10), # Rate at which infectious people shed bacteria to the environment (per day), from Mukandavire et al. (https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3102413/)
            decay_rate = ss.perday(0.033), # Rate at which bacteria in the environment dies (per day), from Chao et al. and Mukandavire et al. citing https://pubmed.ncbi.nlm.nih.gov/8882180/
            p_env_transmit = ss.bernoulli(p=0), # Probability of environmental transmission - filled out later
        )
        self.update_pars(pars, **kwargs)

        # Boolean states
        self.define_states(
            # Susceptible & infected are added automatically, here we add the rest
            ss.BoolState('exposed', label='Exposed'),
            ss.BoolState('symptomatic', label='Symptomatic'),
            ss.BoolState('recovered', label='Recovered'),

            # Timepoint states
            ss.FloatArr('ti_exposed', label='Time of exposure'),
            ss.FloatArr('ti_symptomatic', label='Time of symptoms'),
            ss.FloatArr('ti_recovered', label='Time of recovery'),
            ss.FloatArr('ti_dead', label='Time of death'),
        )
        return

    @property
    def asymptomatic(self):
        return self.infected & ~self.symptomatic

    def init_results(self):
        """ Initialize results """
        super().init_results()
        self.define_results(
            ss.Result('new_deaths', dtype=int,   label='Deaths'),
            ss.Result('cum_deaths', dtype=int,   label='Cumulative deaths'),
            ss.Result('env_prev',   dtype=float, label='Environmental prevalence'),
            ss.Result('env_conc',   dtype=float, label='Environmental concentration'),
        )
        return

    def step_state(self):
        """
        Adapted from https://github.com/optimamodel/gavi-outbreaks/blob/main/stisim/gavi/cholera.py
        Original version by Dom Delport
        """
        # Progress exposed -> infected
        ti = self.ti
        infected = (self.exposed & (self.ti_infected <= ti)).uids
        self.infected[infected] = True
        self.exposed[infected] = False

        # Progress infected -> symptomatic
        symptomatic = (self.infected & (self.ti_symptomatic <= ti)).uids
        self.symptomatic[symptomatic] = True

        # Progress symptomatic -> recovered
        recovered = (self.infectious & (self.ti_recovered <= ti)).uids
        self.exposed[recovered] = False
        self.infected[recovered] = False
        self.symptomatic[recovered] = False
        self.recovered[recovered] = True

        # Trigger deaths
        deaths = (self.ti_dead <= ti).uids
        if len(deaths):
            self.sim.people.request_death(deaths)

        # Update today's environmental prevalence
        self.calc_environmental_prev()

        return

    def calc_environmental_prev(self):
        """
        Calculate environmental prevalence
        """
        p = self.pars
        r = self.results
        ti = self.ti

        n_symptomatic = self.symptomatic.sum()
        n_asymptomatic = self.asymptomatic.sum()
        old_prev = self.results.env_prev[ti-1]

        new_bacteria = (p.shedding_rate * (n_symptomatic + p.asymp_trans * n_asymptomatic)).to_prob()
        old_bacteria = old_prev * np.exp(-p.decay_rate.to_prob())

        r.env_prev[ti] = new_bacteria + old_bacteria
        r.env_conc[ti] = r.env_prev[ti] / (r.env_prev[ti] + p.half_sat_rate)
        return

    def set_prognoses(self, uids, sources=None):
        """ Set prognoses for those who get infected """
        super().set_prognoses(uids, sources)
        ti = self.ti

        self.susceptible[uids] = False
        self.exposed[uids] = True
        self.ti_exposed[uids] = ti

        p = self.pars

        # Determine when exposed become infected
        self.ti_infected[uids] = ti + p.dur_exp2inf.rvs(uids)

        # Determine who becomes symptomatic and when
        symp_uids = p.p_symp.filter(uids)
        self.ti_symptomatic[symp_uids] = self.ti_infected[symp_uids]

        # Determine who dies and when
        dead_uids = p.p_death.filter(symp_uids)
        self.ti_dead[dead_uids] = self.ti_symptomatic[dead_uids] + p.dur_symp2dead.rvs(dead_uids)
        symp_rev_uids = np.setdiff1d(symp_uids, dead_uids)
        asymp_uids = np.setdiff1d(uids, symp_uids)

        # Determine when agents recover
        self.ti_recovered[symp_rev_uids] = self.ti_exposed[symp_rev_uids] + p.dur_symp2rec.rvs(symp_rev_uids)
        self.ti_recovered[asymp_uids] = self.ti_exposed[asymp_uids] + p.dur_asymp2rec.rvs(asymp_uids)

        return

    def infect(self):
        """ Add indirect transmission """
        # Make new cases via direct transmission
        new_cases, sources, networks = super().infect()

        # Make new cases via indirect transmission
        pars = self.pars
        res = self.results
        p_transmit = (res.env_conc[self.ti] * pars.beta_env).to_prob()
        pars.p_env_transmit.set(p=p_transmit)
        new_cases_env = pars.p_env_transmit.filter(self.susceptible)

        if len(new_cases_env):
            new_cases = new_cases + new_cases_env
            sources = np.concatenate([sources, np.full_like(new_cases_env, fill_value=np.nan)])
            networks = np.concatenate([networks, np.full_like(new_cases_env, fill_value=np.nan)])

        return new_cases, sources, networks

    def step_die(self, uids):
        """ Reset infected/recovered flags for dead agents """
        for state in ['susceptible', 'exposed', 'infected', 'symptomatic', 'recovered']:
            self.state_dict[state][uids] = False
        return

    def update_results(self):
        super().update_results()
        res = self.results
        ti = self.ti
        res.new_deaths[ti] = np.count_nonzero(self.ti_dead == ti)
        res.cum_deaths[ti] = np.sum(res.new_deaths[:ti+1])
        return



"""
Define options for Starsim.
All options should be set using set() or directly, e.g.:

    ss.options(verbose=False)
"""
import numpy as np
import numba as nb
import sciris as sc
import matplotlib.font_manager as fm
import starsim as ss # Only used to set references

__all__ = ['dtypes', 'options', 'style']

# Define Starsim-default data types
dtypes = sc.dictobj(
    bool = bool,
    int = np.int64, # np.int32 is actually much slower for indexing arrays
    nbint = nb.int64,
    rand_int = np.int32,
    rand_uint = np.uint32,
    float = np.float32,
    nbfloat = nb.float32,
    result_float = np.float64,
    int_nan = -999999, # Arbitrary placeholder value
)

# Define simple plotting options -- similar to Matplotlib default
rc_starsim = {
    'font.family':       'sans-serif', # Replaced with Mulish in load_fonts() if import succeeds
    'axes.axisbelow':    True, # So grids show up behind
    'axes.spines.right': False,
    'axes.spines.top':   False,
    'legend.frameon':    False,
    'figure.facecolor':  'white',
    'axes.facecolor':    'white',
    'axes.grid':         True,
    'grid.color':        '#f2f2f2',
    'grid.linewidth':    1,
    'lines.linewidth':   2.5,      # Slightly thicker lines
}


# Not public to avoid confusion with ss.options
class Options(sc.objdict):
    """
    Set options for Starsim.

    Use `ss.options.set('defaults')` to reset all values to default, or `ss.options.set(dpi='default')`
    to reset one parameter to default. See `ss.options.help(detailed=True)` for
    more information.

    Options can also be saved and loaded using `ss.options.save()` and `ss.options.load()`.
    See `ss.options.context()` and `ss.options.with_style()` to set options
    temporarily.

    Common options are (see also `ss.options.help(detailed=True)`):

        - verbose:        default verbosity for simulations to use
        - warnings:       how to handle warnings (e.g. print, raise as errors, ignore)

    **Examples**:

        ss.options(verbose=True) # Set more verbosity
        ss.options(warnings='error') # Be more strict about warnings
    """

    def __init__(self):
        super().__init__()
        optdesc, options = self.get_orig_options()  # Get the options
        self.update(options)  # Update this object with them
        self.setattribute('optdesc', optdesc)  # Set the description as an attribute, not a dict entry
        self.setattribute('orig_options', sc.dcp(options))  # Copy the default options
        self.setattribute('_locked', True)
        return

    @staticmethod
    def get_orig_options():
        """
        Set the default options for Starsim -- not to be called by the user, use
        `ss.options.set('defaults')` instead.
        """

        # Options acts like a class, but is actually an objdict for simplicity
        optdesc = sc.objdict()  # Help for the options
        options = sc.objdict()  # The options

        optdesc.verbose = 'Set default level of verbosity (i.e. logging detail): e.g., 0.1 is an update every 10 simulated timesteps.'
        options.verbose = sc.parse_env('STARSIM_VERBOSE', 0.1, float)

        optdesc.license = 'Whether to print the license on import'
        options.license = sc.parse_env('STARSIM_LICENSE', False, bool)

        optdesc.warnings = 'How warnings are handled: options are "warn" (default), "print", and "error"'
        options.warnings = sc.parse_env('STARSIM_WARNINGS', 'warn', str)

        optdesc.warn_convert = 'Whether to warn about automatic time conversions'
        options.warn_convert = sc.parse_env('STARSIM_WARN_CONVERT', True, bool)

        optdesc.check_method_calls = 'How to handle missing required method calls: options are "warn" (default), "die", and "" (False)'
        options.check_method_calls = sc.parse_env('STARSIM_CHECK_METHOD_CALLS', 'warn', str)

        optdesc.time_eps = 'Set size of smallest possible time unit (in units of sim time, e.g. "year" or "day")'
        options.time_eps = sc.parse_env('STARSIM_TIME_EPS', 1e-6, float) # If unit = 'year', corresponds to ~30 seconds

        optdesc.sep = 'Set thousands seperator for text output'
        options.sep = sc.parse_env('STARSIM_SEP', ',', str)

        optdesc.date_sep = 'Set seperator for dates'
        options.date_sep = sc.parse_env('STARSIM_DATE_SEP', '.', str)

        optdesc.jupyter = 'Set whether to use Jupyter settings: -1=auto, 0=False, 1=True'
        options.jupyter = sc.parse_env('STARSIM_JUPYTER', -1, int)

        optdesc.style = 'Set the plotting style: choices are "starsim", "fancy", "simple", or any of Matplotlib\'s'
        options.style = sc.parse_env('STARSIM_STYLE', 'starsim', str)

        optdesc.install_fonts = 'Choose whether or not to install Starsim-specific fonts on first import -- NOTE, only has effect via the environment variable since executed on load'
        options.install_fonts = sc.parse_env('STARSIM_INSTALL_FONTS', True, bool)

        optdesc.show = 'Whether to show the plot immediately (i.e. call plt.show())'
        options.show = sc.parse_env('STARSIM_SHOW', True, bool)

        optdesc.reticulate = 'Set whether to use Reticulate (R) settings'
        options.reticulate = sc.parse_env('STARSIM_RETICULATE', False, bool)

        optdesc.precision = 'Set arithmetic precision'
        options.precision = sc.parse_env('STARSIM_PRECISION', 64, int)

        optdesc.numba_indexing = 'Threshold for the number of indices at which to switch to using Numba (rather than NumPy) for indexing arrays'
        options.numba_indexing = sc.parse_env('STARSIM_NUMBA_INDEXING', 5000, int) # See https://github.com/starsimhub/starsim/issues/1005 for details

        optdesc.single_rng = 'If True, revert to single centralized random number generator like what other agent-based models typically use (not advised; for testing/comparison only.'
        options.single_rng = sc.parse_env('STARSIM_SINGLE_RNG', False, bool)

        return optdesc, options

    def __call__(self, *args, **kwargs):
        """Allow `ss.options(dpi=150)` instead of `ss.options.set(dpi=150)` """
        return self.set(*args, **kwargs)

    def __setitem__(self, key, value):
        """ Do not allow accidental modifications """

        try:
            assert self.getattribute('_locked') # This handles False, not present, etc.
            locked = True
        except:
            locked = False

        if locked:
            if key in self:
                self.set(key=key, value=value)
            else:
                errormsg = f'Cannot set option "{key}"; valid options are:\n{sc.newlinejoin(sorted(self.keys()))}\n\nSee ss.options.disp() for details.'
                raise sc.KeyNotFoundError(errormsg)
        else:
            return super().__setitem__(key, value)

    def to_dict(self):
        ''' Pull out only the settings from the options object '''
        return {k:v for k,v in self.items()}

    def __repr__(self):
        """ Brief representation """
        output = sc.objectid(self)
        output += 'Starsim options (see also ss.options.disp()):\n'
        output += sc.pp(self.to_dict(), output=True)
        return output

    def __enter__(self):
        """ Allow to be used in a with block """
        return self

    def __exit__(self, *args, **kwargs):
        """ Allow to be used in a with block """
        try:
            reset = {}
            for k, v in self.on_entry.items():
                if self[k] != v:  # Only reset settings that have changed
                    reset[k] = v
            self.set(**reset)
            self.delattribute('on_entry')
        except AttributeError as E:
            errormsg = 'Please use ss.options.context() if using a with block'
            raise AttributeError(errormsg) from E
        return

    def disp(self):
        """ Detailed representation """
        output = 'Starsim options (see also ss.options.help()):\n'
        keylen = 10  # Maximum key length  -- "interactive"
        for k, v in self.items():
            keystr = sc.colorize(f'  {k:>{keylen}s}: ', fg='cyan', output=True)
            reprstr = sc.pp(v, output=True)
            reprstr = sc.indent(n=keylen + 4, text=reprstr, width=None)
            output += f'{keystr}{reprstr}'
        print(output)
        return

    def help(self, *args, detailed=False, output=False):
        """
        Print information about options.

        Args:
            *args (list): if provided, print out detailed help on one or more specific options
            detailed (bool): whether to print out full help
            output (bool): whether to return a list of the options

        **Example**:

            ss.options.help(detailed=True)
        """
        # If not detailed, just print the docstring for sc.options
        if not detailed and not len(args):
            print(self.__doc__)
            return

        n = 15 # Size of indent
        optdict = sc.objdict()
        orig_keys = self.orig_options.keys()
        for key in orig_keys:
            entry = sc.objdict()
            entry.key = key
            entry.current = sc.indent(n=n, width=None, text=sc.pp(self[key], output=True)).rstrip()
            entry.default = sc.indent(n=n, width=None, text=sc.pp(self.orig_options[key], output=True)).rstrip()
            entry.variable = f'STARSIM_{key.upper()}' # NB, hard-coded above!
            entry.desc = sc.indent(n=n, text=self.optdesc[key])
            optdict[key] = entry

        if len(args):
            try:
                optdict = sc.objdict({k:optdict[k] for k in args})
            except Exception as e:
                errormsg = f'Cannot provide help on invalid setting(s) "{sc.strjoin(args)}"; valid keys are:\n{sc.newlinejoin(sorted(orig_keys))}'
                raise sc.KeyNotFoundError(errormsg) from e

        print('Starsim global options ("Environment" = name of corresponding environment variable):')
        for k, key, entry in optdict.enumitems():
            sc.heading(f'{k}. {key}', spaces=0, spacesafter=0)
            changestr = '' if entry.current == entry.default else ' (modified)'
            print(f'          Key: {key}')
            print(f'      Current: {entry.current}{changestr}')
            print(f'      Default: {entry.default}')
            print(f'  Environment: {entry.variable}')
            print(f'  Description: {entry.desc}')

        if detailed:
            sc.heading('Methods:', spacesafter=0)
            print("""
    ss.options(key=value) -- set key to value
    ss.options[key] -- get or set key
    ss.options.set() -- set option(s)
    ss.options.get_default() -- get default setting(s)
    ss.options.set_style() -- change the style for plotting (can also use ss.style() in a context block)
""")

        if output:
            return optdict
        else:
            return

    def set(self, key=None, value=None, use=False, **kwargs):
        """
        Actually change the style. See `ss.options.help()` for more information.

        Args:
            key    (str):    the parameter to modify, or 'defaults' to reset everything to default
            value  (varies): the value to specify; use None or 'default' to reset to default
            use (bool): whether to use the chosen style
            kwargs (dict):   if supplied, set multiple key-value pairs

        **Example**:
            ss.options.set(dpi=50) # Equivalent to ss.options(dpi=50)
        """

        # Reset to defaults
        if key in ['default', 'defaults']:
            kwargs = self.orig_options  # Reset everything to default

        # Handle other keys
        elif key is not None:
            kwargs = sc.mergedicts(kwargs, {key: value})

        # Reset options
        for key, value in kwargs.items():
            if key not in self:
                keylist = self.orig_options.keys()
                keys = '\n'.join(keylist)
                errormsg = f'Option "{key}" not recognized; options are "defaults" or:\n{keys}\n\nSee help(ss.options.set) for more information.'
                raise sc.KeyNotFoundError(errormsg)
            else:
                if value in [None, 'default']:
                    value = self.orig_options[key]
                super().__setitem__(key, value) # Needed since we overwrite __setitem__ to call this

                # Handle special cases
                if key == 'jupyter':
                    self.set_jupyter()
                elif key == 'precision':
                    self.set_precision()
                elif key == 'numba_indexing':
                    self.refresh_references()
                elif key == 'style':
                    self.set_style()

        return

    def context(self, **kwargs):
        """
        Alias to set(), for use in a "with" block.

        **Examples**:

            # Silence all output
            with ss.options.context(verbose=0):
                ss.Sim().run()

            # Convert warnings to errors
            with ss.options.context(warnings='error'):
                ss.Sim(location='not a location').init()

            # Use with_style(), not context(), for plotting options
            with ss.options.with_style(dpi=50):
                ss.Sim().run().plot()
        """

        # Store current settings
        on_entry = {k: self[k] for k in kwargs.keys()}
        self.setattribute('on_entry', on_entry)

        # Make changes
        self.set(**kwargs)
        return self

    def get_default(self, key):
        """ Helper function to get the original default options """
        return self.orig_options[key]

    def changed(self, key):
        """ Check if current setting has been changed from default """
        if key in self.orig_options:
            return self[key] != self.orig_options[key]
        else:
            return None

    @property
    def is_jupyter(self):
        """ Check if we're in an Jupyter environment """
        return [False, True, sc.isjupyter()][self.jupyter] # 0 = False, 1 = True, -1 = sc.isjupyter()

    def set_jupyter(self):
        """ Set Sciris' Jupyter settings as well """
        sc.options(jupyter=self.is_jupyter)
        return

    def set_precision(self):
        """ Change the arithmetic precision used by Starsim/NumPy """
        if self.precision == 32:
            dtypes.update(dict(
                int       = np.int64, # Never set this or result_float to 32-bit since messed up array indexing
                rand_int  = np.int32,
                rand_uint = np.uint32,
                float     = np.float32,
                nbfloat   = nb.float32,
            ))
        elif self.precision == 64:
            dtypes.update(dict(
                int       = np.int64,
                rand_int  = np.int64,
                rand_uint = np.uint64,
                float     = np.float64,
                nbfloat   = nb.float64,
            ))
        else:
            errormsg = f'Precision {self.precision} not recognized; must be 32 or 64'
            raise ValueError(errormsg)
        self.refresh_references()
        return

    def refresh_references(self):
        """ Many modules copy settings into them for speed; here we refresh those """
        files = ['arrays', 'demographics', 'diseases', 'distributions', 'networks']

        # Update Numba threshold
        ss.arrays.numba_indexing = self.numba_indexing

        # Update precision
        for file in files:
            ss_file = getattr(ss, file) # e.g. ss.arrays
            for key,value in dtypes.items():
                ss_key = f'ss_{key}' # e.g. int -> ss_int
                if hasattr(ss_file, ss_key): # e.g. ss.arrays.ss_int
                    setattr(ss_file, ss_key, value)
        return

    def set_style(self):
        """ Change the plotting style """
        if self.style == 'starsim':
            style = sc.dcp(rc_starsim)
        else:
            style = self.style

        # Since Options is locked
        self.setattribute('_style', style)
        return



def load_fonts(folder=None, name='Mulish', rebuild=False, verbose=False, **kwargs):
    """
    Helper function to load custom fonts for plotting -- (usually) not for the user.

    Note: if fonts don't load, try running ``ss.settings.load_fonts(rebuild=True)``,
    and/or rebooting the system.

    Args:
        folder (str): the folder to add fonts from
        name (str): the name of the font to load
        rebuild (bool): whether to rebuild the font cache
        verbose (bool): whether to print out progress/errors
    """
    if folder is None:
        folder = str(sc.thispath(__file__) / 'assets')
    sc.fonts(add=folder, rebuild=rebuild, verbose=verbose, **kwargs)

    # Try to find the font, and if it succeeds, update the styles
    try:
        fm.findfont(name, fallback_to_default=False) # Raise an exception if the font isn't found
        rc_starsim['font.family']  = name # Need to set both
        if verbose: print(f'Default Starsim font reset to "{name}"')
    except Exception as E:
        if verbose: print(f'Could not find font {name}: {str(E)}')
    return


# Create the options on module load
options = Options()

# Load the fonts
if options.install_fonts:
    load_fonts()

# Set the style
options.set_style()


def style(style=None, **kwargs):
    """
    Set the style in a with block.

    Note: Starsim comes bundled with three fonts: Mulish (default), Raleway, and Rosario.
    Use `font='sans-serif'` to use the Matplotlib default font.

    Args:
        style (str): the style to use; if None, use current; otherwise, 'starsim', 'simple', 'fancy', plus all of the Matplotlib styles are options
        **kwargs (dict): passed to `sc.options.with_style()`

    **Examples**::

        # Create a plot using default Starsim styling
        with ss.style():
            plt.plot()

        # Create a plot using a built-in Matplotlib style
        with ss.style('seaborn-v0_8-whitegrid'):
            plt.plot()

        # Customize the current style
        with ss.style(font='Rosario'):
            plt.plot()
    """
    if style is None:
        style = options._style
    elif style in ['starsim', 'default']:
        style = sc.dcp(rc_starsim)

    # Rename to avoid name collisions (e.g. font vs font.family)
    mapping = {'font':'font.family', 'facecolor':'axes.facecolor', 'grid':'axes.grid'}
    for key,val in mapping.items():
        if key in kwargs:
            kwargs[val] = kwargs.pop(key)

    # Reset
    out = sc.options.with_style(style, **kwargs)
    return out


"""
Test Starsim arrays (ss.Arr, ss.FloatArr, etc)
"""
import numpy as np
import sciris as sc
import starsim as ss
import matplotlib.pyplot as plt
import pytest

sc.options.interactive = False # Assume not running interactively
# ss.options.warnings = 'error' # For additional debugging

small = 100
medium = 1000

# %% Define the tests

@sc.timer()
def test_arrs():
    sc.heading('Testing Arr objects')
    o = sc.objdict()

    # Create a sim with only births
    pars = dict(n_agents=medium, diseases='sis', networks='random')
    p1 = sc.mergedicts(pars, birth_rate=ss.freqperyear(10))
    p2 = sc.mergedicts(pars, death_rate=ss.freqperyear(10))
    s1 = ss.Sim(pars=p1).run()
    s2 = ss.Sim(pars=p2).run()

    # Tests
    assert len(s1.people.auids) > len(s2.people.auids), 'Should have more people with births'
    assert np.array_equal(s1.people.age, s1.people.age.raw[s1.people.auids]), 'Different ways of indexing age should match'
    assert np.array_equal(s1.people.alive.uids, s1.people.auids), 'Active/alive agents should match'
    assert np.all(s2.people.alive), 'All agents should be alive when indexed like this'
    assert not np.all(s2.people.alive.raw), 'Some agents should not be alive when indexed like this'
    assert np.array_equal(~s1.people.female, s1.people.male), 'Definition of men does not match'
    assert isinstance(s1.people.age < 5, ss.BoolArr), 'Performing logical operations should return a BoolArr'
    assert np.array_equal(s1.people.age[s1.people.age < 5],s1.people.age[s1.people.age.values < 5])

    # Test BoolArr & uids operations
    test_bool = s1.people.age > 20
    test_uids = test_bool.uids

    # Standard operations
    assert np.all(test_bool == test_uids)
    assert np.all(~(test_bool != test_uids))
    assert np.array_equal(s1.people.female & test_bool, s1.people.female & test_uids)
    assert np.array_equal(s1.people.female | test_bool, s1.people.female | test_uids)
    assert np.array_equal(s1.people.female ^ test_bool, s1.people.female ^ test_uids)
    assert np.array_equal(test_uids & s1.people.female,  test_uids & s1.people.female.uids)
    assert np.array_equal(test_uids | s1.people.female,  test_uids | s1.people.female.uids)
    assert np.array_equal(test_uids ^ s1.people.female,  test_uids ^ s1.people.female.uids)

    # Inplace BoolArr operations
    a = s1.people.female & test_bool
    b = sc.dcp(s1.people.female)
    original_id = id(b)
    original_id_raw = id(b.raw)
    b &= test_bool
    assert np.array_equal(a, b)
    assert id(b) == original_id
    assert id(b.raw) == original_id_raw

    a = s1.people.female & test_uids
    b = sc.dcp(s1.people.female)
    original_id = id(b)
    original_id_raw = id(b.raw)
    b &= test_uids
    assert np.array_equal(a, b)
    assert id(b) == original_id
    assert id(b.raw) == original_id_raw

    a = s1.people.female | test_bool
    b = sc.dcp(s1.people.female)
    original_id = id(b)
    original_id_raw = id(b.raw)
    b |= test_bool
    assert np.array_equal(a, b)
    assert id(b) == original_id
    assert id(b.raw) == original_id_raw

    a = s1.people.female ^ test_uids
    b = sc.dcp(s1.people.female)
    original_id = id(b)
    original_id_raw = id(b.raw)
    b ^= test_uids
    assert np.array_equal(a, b)
    assert id(b) == original_id
    assert id(b.raw) == original_id_raw

    o.s1 = s1
    o.s2 = s2

    return o


@sc.timer()
def test_arr_inplace():
    # Note that the test function above focusses on BoolArr specifially which has its own version of some of the operators
    sc.heading('Testing Arr in-place vs non-in-place arithmetic operators')

    # Create two synthetic FloatArrs without needing a full sim
    a = ss.FloatArr('a', default=3.0, mock=10)
    b = ss.FloatArr('b', default=2.0, mock=10)

    # --- In-place: += must update values without replacing the Arr or .raw array ---
    original_id     = id(a)
    original_id_raw = id(a.raw)
    expected        = a.raw.copy() + b.raw.copy()
    a += b
    assert np.allclose(a.raw, expected),       'In-place += must produce correct values'
    assert id(a)     == original_id,           'In-place += must not replace the Arr object'
    assert id(a.raw) == original_id_raw,       'In-place += must not replace the .raw array'

    # --- Non-in-place: + must return an independent copy without modifying the original ---
    original_raw = a.raw.copy()
    c = a + b
    assert id(c)     != id(a),                 'Non-in-place + must return a new Arr object'
    assert id(c.raw) != id(a.raw),             'Non-in-place + must use a new .raw array'
    assert np.allclose(a.raw, original_raw),   'Non-in-place + must not modify the original Arr'
    assert np.allclose(c.raw, original_raw + b.raw), 'Non-in-place + must produce correct values'

    masked_auids = ss.uids(np.array([0, 1, 2, 4, 5, 6, 7, 8, 9], dtype=int))
    inactive_uid = 3

    # --- In-place masked Arr += Arr ---
    a = ss.FloatArr('a', default=3.0, mock=10)
    b = ss.FloatArr('b', raw=np.arange(10, dtype=float), mock=10)
    a.people.auids = masked_auids
    b.people.auids = masked_auids

    original_id = id(a)
    original_id_raw = id(a.raw)
    expected = a.raw.copy()
    expected[masked_auids] += b.raw[masked_auids]

    a += b
    assert np.allclose(a.raw, expected), 'In-place += must update raw values for active UIDs only'
    assert a.raw[inactive_uid] == 3.0,   'In-place += must not modify inactive UIDs'
    assert id(a) == original_id,         'In-place += must not replace the Arr object'
    assert id(a.raw) == original_id_raw, 'In-place += must not replace the .raw array'

    # --- In-place masked Arr -= Arr ---
    a = ss.FloatArr('a', default=3.0, mock=10)
    b = ss.FloatArr('b', raw=np.arange(10, dtype=float), mock=10)
    a.people.auids = masked_auids
    b.people.auids = masked_auids
    expected = a.raw.copy()
    expected[masked_auids] -= b.raw[masked_auids]
    a -= b
    assert np.allclose(a.raw, expected), 'In-place -= must update raw values for active UIDs only'
    assert a.raw[inactive_uid] == 3.0,   'In-place -= must not modify inactive UIDs'

    # --- In-place masked Arr +/- scalar ---
    a = ss.FloatArr('a', default=3.0, mock=10)
    a.people.auids = masked_auids
    expected = a.raw.copy()
    expected[masked_auids] += 2
    a += 2
    assert np.allclose(a.raw, expected), 'In-place += scalar must update active UIDs only'
    expected[masked_auids] -= 1
    a -= 1
    assert np.allclose(a.raw, expected), 'In-place -= scalar must update active UIDs only'
    assert a.raw[inactive_uid] == 3.0,   'Scalar in-place arithmetic must not modify inactive UIDs'

    # --- In-place masked Arr += ndarray aligned to active UIDs ---
    a = ss.FloatArr('a', default=3.0, mock=10)
    a.people.auids = masked_auids
    rhs = np.arange(len(masked_auids), dtype=float)
    expected = a.raw.copy()
    expected[masked_auids] += rhs
    a += rhs
    assert np.allclose(a.raw, expected), 'In-place += ndarray must treat the ndarray as aligned to active UIDs'

    # --- In-place masked Arr += ndarray with mismatched length ---
    a = ss.FloatArr('a', default=3.0, mock=10)
    a.people.auids = masked_auids
    try:
        a += np.arange(len(masked_auids) + 1, dtype=float)
        raise AssertionError('In-place += with a mismatched ndarray length must raise an error')
    except ValueError:
        pass

    # --- In-place masked Arr += Arr with different active masks ---
    a = ss.FloatArr('a', default=3.0, mock=10)
    b = ss.FloatArr('b', raw=np.arange(10, dtype=float), mock=10)
    other_masked_auids = ss.uids(np.array([0, 1, 2, 3, 4, 5, 6, 7, 8], dtype=int))
    a.people.auids = masked_auids
    b.people.auids = other_masked_auids
    expected = a.raw.copy()
    expected[masked_auids] += b.raw[masked_auids]
    a += b
    assert np.allclose(a.raw, expected), 'Arr-to-Arr arithmetic must read RHS values from other.raw[self.auids]'
    assert a.raw[inactive_uid] == 3.0,   'Mismatched RHS masks must not modify inactive UIDs'

    # --- Non-in-place masked Arr + Arr ---
    a = ss.FloatArr('a', default=3.0, mock=10)
    b = ss.FloatArr('b', raw=np.arange(10, dtype=float), mock=10)
    a.people.auids = masked_auids
    b.people.auids = masked_auids
    original_raw = a.raw.copy()
    c = a + b
    expected = original_raw.copy()
    expected[masked_auids] += b.raw[masked_auids]
    assert id(c) != id(a),               'Masked non-in-place + must return a new Arr object'
    assert id(c.raw) != id(a.raw),       'Masked non-in-place + must use a new .raw array'
    assert np.allclose(a.raw, original_raw), 'Masked non-in-place + must not modify the original Arr'
    assert np.allclose(c.raw, expected), 'Masked non-in-place + must preserve inactive raw values and update active UIDs'
    assert c.raw[inactive_uid] == original_raw[inactive_uid], 'Masked non-in-place + must preserve inactive raw values'

    return a, b


@sc.timer()
def test_arr_type_conversion():
    sc.heading('Testing Arr type conversion and mixed bool/float arithmetic')

    a = np.array([True, False], dtype=bool)
    b = np.array([3, 4], dtype=float)
    c = ss.BoolArr('c', raw=np.array([True, False], dtype=bool), mock=2)
    d = ss.FloatArr('d', raw=np.array([3, 4], dtype=float), mock=2)

    ab = a * b
    ba = b * a
    assert ab.dtype == float
    assert ba.dtype == float
    assert np.array_equal(ab, np.array([3.0, 0.0]))
    assert np.array_equal(ba, np.array([3.0, 0.0]))

    with pytest.raises(TypeError):
        a_copy = a.copy()
        a_copy *= b

    b_copy = b.copy()
    b_copy *= a
    assert b_copy.dtype == float
    assert np.array_equal(b_copy, np.array([3.0, 0.0]))

    cd = c * d
    dc = d * c
    assert isinstance(cd, ss.FloatArr)
    assert isinstance(dc, ss.FloatArr)
    assert cd.raw.dtype == float
    assert dc.raw.dtype == float
    assert np.array_equal(cd.raw, np.array([3.0, 0.0]))
    assert np.array_equal(dc.raw, np.array([3.0, 0.0]))

    with pytest.raises(TypeError):
        c_copy = ss.BoolArr('c_copy', raw=np.array([True, False], dtype=bool), mock=2)
        c_copy *= d

    d_copy = ss.FloatArr('d_copy', raw=np.array([3, 4], dtype=float), mock=2)
    d_copy *= c
    assert d_copy.raw.dtype == float
    assert np.array_equal(d_copy.raw, np.array([3.0, 0.0]))

    return dict(ab=ab, ba=ba, cd=cd, dc=dc)


# %% Run as a script
if __name__ == '__main__':
    do_plot = True
    sc.options(interactive=do_plot)

    # Start timing
    T = sc.tic()

    # Run tests
    sims  = test_arrs()
    arrs  = test_arr_inplace()
    vals  = test_arr_type_conversion()

    sc.toc(T)
    plt.show()
    print('Done.')



#!/usr/bin/env python3
"""
Benchmark a few representative simulation scenarios.

Examples:
    ./benchmark_examples.py
    ./benchmark_examples.py repeats=2
    ./benchmark_examples.py case=complex_30
"""
import sciris as sc
import starsim as ss
import starsim_examples as sse

sc.options(interactive=False)


def make_simple_setup():
    """A tiny run where setup costs dominate the total runtime."""
    return ss.Sim(
        n_agents=50_000,
        dur=1,
        dt=1,
        rand_seed=1,
        verbose=0,
        diseases=ss.SIS(),
        networks=ss.RandomNet(),
        label='simple_setup',
    )


def make_complex_30():
    """A shorter run with substantially more model structure."""
    hiv = sse.HIV()
    hiv.pars['beta'] = {'mf': [0.15, 0.10], 'maternal': [0.2, 0], 'random': [0, 0]}

    return ss.Sim(
        n_agents=80_000,
        dur=15,
        dt=0.5,  # 30 timesteps
        rand_seed=1,
        verbose=0,
        diseases=[ss.SIS(), hiv],
        networks=[ss.RandomNet(), ss.MFNet(), ss.MaternalNet()],
        demographics=ss.Pregnancy(),
        label='complex_30',
    )


def make_moderate():
    """A moderately complex run pushed by many timesteps."""
    return ss.Sim(
        n_agents=45_000,
        dur=30,
        dt=0.1,  # 300 timesteps
        rand_seed=1,
        verbose=0,
        diseases=[ss.SIR(), ss.SIS()],
        networks=[ss.RandomNet(), ss.MFNet()],
        demographics=True,
        label='moderate_300',
    )


CASES = sc.objdict(
    simple_setup=sc.objdict(
        label='Simple setup-dominated simulation',
        timesteps=1,
        repeats=7,
        factory=make_simple_setup,
    ),
    complex=sc.objdict(
        label='High-complexity simulation with 30 timesteps',
        timesteps=30,
        repeats=2,
        factory=make_complex_30,
    ),
    moderate=sc.objdict(
        label='Moderate-complexity simulation with 300 timesteps',
        timesteps=100,
        repeats=3,
        factory=make_moderate,
    ),
)


def benchmark_case(name, case, repeats=None):
    repeats = sc.ifelse(repeats, case.repeats)
    init_times = []
    run_times = []

    for r in range(repeats):
        print(f'  Repeat {r + 1}/{repeats}')

        sim = case.factory()

        t0 = sc.tic()
        sim.init()
        init_times.append(sc.toc(t0, output=True))

        t0 = sc.tic()
        sim.run()
        run_times.append(sc.toc(t0, output=True))

    row = sc.objdict(
        case=name,
        label=case.label,
        n_agents=int(sim.pars.n_agents),
        timesteps=case.timesteps,
        init_best=round(min(init_times), 3),
        init_mean=round(sum(init_times) / len(init_times), 3),
        run_best=round(min(run_times), 3),
        run_mean=round(sum(run_times) / len(run_times), 3),
        total_best=round(min(i + r for i, r in zip(init_times, run_times)), 3),
        total_mean=round((sum(init_times) + sum(run_times)) / repeats, 3),
    )
    return row


def get_cases(selection):
    if selection in [None, 'all']:
        return CASES
    if selection not in CASES:
        errormsg = f'Invalid case "{selection}"; choices are: {sc.strjoin(CASES.keys())}, all'
        raise ValueError(errormsg)
    return sc.objdict({selection: CASES[selection]})


def benchmark_examples(repeats=None, case='all'):
    selected = get_cases(case)
    rows = []

    for name, spec in selected.items():
        sc.heading(f'Benchmarking: {spec.label}')
        rows.append(benchmark_case(name, spec, repeats=repeats))
        print()

    df = sc.dataframe(rows)
    df = df[['case', 'n_agents', 'timesteps', 'init_best', 'init_mean', 'run_best', 'run_mean', 'total_best', 'total_mean']]
    print('Summary (seconds):')
    print(df.round(3).to_string(index=False))
    return df


if __name__ == '__main__':
    args = sc.argparse(repeats=None, case='all')
    benchmark_examples(repeats=args.repeats, case=args.case)



"""
Test Starsim features not covered by other test files
"""
import numpy as np
import sciris as sc
import starsim as ss
import starsim_examples as sse
import matplotlib.pyplot as plt

sc.options.interactive = False # Assume not running interactively
# ss.options.warnings = 'error' # For additional debugging

small = 100
medium = 1000

# %% Define the tests

@sc.timer()
def test_microsim(do_plot=False):
    sc.heading('Test small HIV simulation')

    # Make HIV module
    hiv = sse.HIV()
    # Set beta. The first entry represents transmission risk from infected p1 -> susceptible p2
    # Need to be careful to get the ordering right. The set-up here assumes that in the simple
    # sexual  network, p1 is male and p2 is female. In the maternal network, p1=mothers, p2=babies.
    hiv.pars['beta'] = {'mf': [0.15, 0.10], 'maternal': [0.2, 0]}

    sim = ss.Sim(
        people=ss.People(small),
        networks=[ss.MFNet(), ss.MaternalNet()],
        demographics=ss.Pregnancy(),
        diseases=hiv,
        copy_inputs = False, # So we can reuse hiv
    )
    sim.init()
    sim.run()

    if do_plot:
        plt.figure()
        plt.plot(hiv.timevec, hiv.results.n_infected)
        plt.title('HIV number of infections')

    return sim


@sc.timer()
def test_results():
    sc.heading('Testing results export and plotting')

    # Make a sim with 2 SIS models with varying units and dt
    d1 = ss.SIS(dt=ss.years(1/12), name='sis1')
    d2 = ss.SIS(dt=ss.years(0.5), name='sis2')
    sim = ss.Sim(n_agents=medium, diseases=[d1, d2], networks='random')

    # Run sim and pull out disease results
    sim.run()
    rs1 = sim.results.sis1
    rs2 = sim.results.sis2

    # Export a single result to a series or dataframe
    res = rs1.new_infections
    res_df = res.to_df()
    res_series = res.to_series()
    assert res[-1] == res_df.iloc[-1].value == res_series.iloc[-1]

    # Test resampling (always returns a Result)
    resampled = rs1.new_infections.resample(new_unit='year')
    assert isinstance(resampled, ss.Result)
    assert len(resampled.values) < len(rs1.new_infections.values)

    # Test that resample and annualize produce the same values
    annualized = rs1.new_infections.annualize()
    assert np.allclose(resampled.values, annualized.values, rtol=1e-10)

    # Test Results.annualize() (annualizes all results in the group)
    rs1_annual = rs1.annualize()
    assert isinstance(rs1_annual, ss.Results)
    assert len(rs1_annual.new_infections.values) == len(annualized.values)
    assert np.allclose(rs1_annual.new_infections.values, annualized.values, rtol=1e-10)

    # Export results of a whole module to a dataframe
    dfs = rs1.to_df()
    assert res_df.value.sum() == dfs.cum_infections.values[-1] == sim.summary.sis1_cum_infections

    # Export resampled summary of results to dataframe
    dfy1 = rs1.to_df(resample='year')
    dfy2 = rs2.to_df(resample='5YE')
    assert dfs.new_infections.iloc[:12].sum() == dfy1.new_infections.iloc[0]
    assert rs2.n_susceptible[:2].mean() == dfy2.n_susceptible.iloc[0]  # Entries 0 and 1 represent 2000
    assert rs2.n_susceptible[2:12].mean() == dfy2.n_susceptible.iloc[1]  # Entries 2-12 correspond to 2001-2005

    # Export whole sim to unified annualized dataframe
    sim_df = sim.to_df(resample='year', use_years=True)
    assert sim_df.sis1_n_infected.values[0] == rs1.n_infected[:12].mean()
    assert sim_df.sis2_n_infected.values[0] == rs2.n_infected[:2].mean()

    # Plot
    res.plot()
    sim.results.sis1.plot()
    sim.results.sis2.plot()

    return sim


@sc.timer()
def test_deepcopy():
    sc.heading('Testing deepcopy')
    s1 = ss.Sim(pars=dict(diseases='sir', networks=sse.EmbeddingNet()), n_agents=small)
    s1.init()

    s2 = sc.dcp(s1)

    s1.run()
    s2.run()
    s1.plot()
    s2.plot()

    ss.diff_sims(s1, s2, full=True)
    assert np.allclose(s1.summary[:], s2.summary[:], rtol=0, atol=0, equal_nan=True)

    return s1, s2


@sc.timer()
def test_deepcopy_until():
    sc.heading('Testing deepcopy with until')
    s1 = ss.Sim(diseases=ss.SIR(init_prev=0.1), networks='random', n_agents=small)
    s1.init()

    s1.run(until=ss.date(2005))

    s2 = sc.dcp(s1)

    s1.run()
    s2.run()
    s1.plot()
    s2.plot()

    assert np.allclose(s1.summary[:], s2.summary[:], rtol=0, atol=0, equal_nan=True)

    return s1, s2


@sc.timer()
def test_custom_imports():
    sc.heading('Testing custom imports')
    o = sc.objdict()

    class MyDisease(ss.SIS):
        pass

    class MyNetwork(ss.RandomNet):
        pass

    my_modules = [MyDisease, MyNetwork]

    # Make both starsim_examples and custom modules searchable
    ss.register_modules(sse, my_modules)
    sim = ss.Sim(n_agents=1000, diseases=['hiv', 'mydisease'], networks='mynetwork')
    sim.run()

    return o


# %% Run as a script
if __name__ == '__main__':
    do_plot = True
    sc.options(interactive=do_plot)

    # Start timing
    T = sc.tic()

    # Run tests
    sim1  = test_microsim(do_plot)
    sim2  = test_results()
    sims1 = test_deepcopy()
    sims2 = test_deepcopy_until()
    mods  = test_custom_imports()

    sc.toc(T)
    plt.show()
    print('Done.')



"""
Version and license information.
"""

__all__ = ['__version__', '__versiondate__', '__license__']

__version__ = '3.3.0'
__versiondate__ = '2026-03-27'
__license__ = f'Starsim {__version__} ({__versiondate__}) — © 2023-2026 by IDM'



siskw = dict(dur_inf=ss.datedur(days=50), beta=ss.perday(0.01), waning=ss.perday(0.005))
kw = dict(n_agents=1000, start='2001-01-01', stop='2001-07-01', networks='random', copy_inputs=False, verbose=0)

print('Year-year')
sis1 = ss.SIS(dt=1/365, **sc.dcp(siskw))
sim1 = ss.Sim(dt=1/365, diseases=sis1, label='year-year', **kw)

print('Day-day')
sis2 = ss.SIS(dt=ss.days(1), **sc.dcp(siskw))
sim2 = ss.Sim(dt=ss.days(1), diseases=sis2, label='day-day', **kw)

print('Day-year')
sis3 = ss.SIS(dt=ss.days(1), **sc.dcp(siskw))
sim3 = ss.Sim(dt=1/365, diseases=sis3, label='day-year', **kw)

print('Year-day')
sis4 = ss.SIS(dt=1/365, **sc.dcp(siskw))
sim4 = ss.Sim(dt=ss.days(1), diseases=sis4, label='year-day', **kw) 


"""
Test the Dist object from distributions.py
"""

import numpy as np
import sciris as sc
import scipy.stats as sps
import starsim as ss
import matplotlib.pyplot as plt
import pytest

n = 1_000_000 # For testing statistical properties and performance of distributions
m = 5 # For just testing that drawing works
dpy = 365.25 # Number of days per year
sc.options(interactive=False)


def plot_rvs(rvs, times=None, nrows=None):
    fig = plt.figure(figsize=(12,12))
    nrows, ncols = sc.getrowscols(len(rvs), nrows=nrows)
    for i,name,r in rvs.enumitems():
        plt.subplot(nrows, ncols, i+1)
        plt.hist(r.astype(float))
        title = times[name] if times else name
        plt.title(title)
        sc.commaticks()
    sc.figlayout()
    return fig


def make_sim():
    """ Make a tiny sim for initializing the distributions """
    sim = ss.Sim(n_agents=100).init() # Need an empty sim to initialize properly
    return sim


# %% Define the tests
@sc.timer()
def test_dist(m=m):
    """ Test the Dist class """
    sc.heading('Testing the basic Dist call')
    dist = ss.Dist(distname='random', name='test', strict=False)
    dist.init()
    rvs = dist(m)
    print(rvs)
    assert 0 < rvs.min() < 1, 'Values should be between 0 and 1'

    # Test other options
    dist.show_state()
    dist.plot_hist()
    return rvs


@sc.timer()
def test_custom_dists(n=n, do_plot=False):
    """ Test all custom dists """
    sc.heading('Testing all custom distributions')

    o     = sc.objdict()
    dists = sc.objdict()
    rvs   = sc.objdict()
    times = sc.objdict()
    for name in ss.dist_list:
        dist_class = getattr(ss, name)
        dist = dist_class(name='test', strict=False)
        dist.init()
        dists[name] = dist
        sc.tic()
        rvs[name] = dist.rvs(n)
        times[name] = sc.toc(name, unit='ms', output='message')
        print(f'{name:10s}: mean = {rvs[name].mean():n}')

    if do_plot:
        plot_rvs(rvs, times=times, nrows=5)

    o.dists = dists
    o.rvs = rvs
    o.times = times
    return o


@sc.timer()
def test_dists(n=n, do_plot=False):
    """ Test the Dists container """
    sc.heading('Testing Dists container')
    testvals = np.zeros((2,2))

    # Create the objects twice
    for i in range(2):

        # Create a complex object containing various distributions
        obj = sc.prettyobj()
        obj.a = sc.objdict()
        obj.a.mylist = [ss.random(), ss.Dist(distname='uniform', low=2, high=3)]
        obj.b = dict(d3=ss.weibull(c=2), d4=ss.constant(v=0.3))
        dists = ss.Dists(obj).init(sim=make_sim())

        # Call each distribution twice
        for j in range(2):
            rvs = sc.objdict()
            for key,dist in dists.dists.items():
                rvs[str(dist)] = dist(n)
                dist.jump() # Reset
            testvals[i,j] = rvs[0][283] # Pick one of the random values and store it

    # Check that results are as expected
    print(testvals)
    assert np.all(testvals[0,:] == testvals[1,:]), 'Newly initialized objects should match'
    assert np.all(testvals[:,0] != testvals[:,1]), 'After jumping, values should be different'

    if do_plot:
        plot_rvs(rvs)

    o = sc.objdict()
    o.dists = dists
    o.rvs = rvs
    return dists


@sc.timer()
def test_scipy(m=m):
    """ Test that SciPy distributions also work """
    sc.heading('Testing SciPy distributions')

    # Make SciPy distributions in two different ways
    dist1 = ss.Dist(dist=sps.expon, name='scipy', scale=2, strict=False).init() # Version 1: callable
    dist2 = ss.Dist(dist=sps.expon(scale=2), name='scipy', strict=False).init() # Version 2: frozen
    rvs1 = dist1(m)
    rvs2 = dist2(m)

    # Check that they match
    print(rvs1)
    assert np.array_equal(rvs1, rvs2), 'Arrays should match'

    return dist1, dist2


@sc.timer()
def test_exceptions(m=m):
    """ Check that exceptions are being appropriately raised """
    sc.heading('Testing exceptions and strict')

    # Create a strict distribution
    dist = ss.random(strict=True, auto=False)
    with pytest.raises(ss.distributions.DistNotInitializedError):
        dist(m) # Check that we can't call an uninitialized

    # Initialize and check we can't call repeatedly
    dist.init(trace='test', sim=make_sim())
    rvs = dist(m)
    with pytest.raises(ss.distributions.DistNotReadyError):
        dist(m) # Check that we can't call an already-used distribution

    # Check that we can with a non-strict Dist
    dist2 = ss.random(strict=False)
    dist2.init(trace='test')
    rvs2 = sc.autolist()
    for i in range(2):
        rvs2 += dist2(m) # We should be able to call multiple times with no problem

    print(rvs)
    print(rvs2)
    assert np.array_equal(rvs, rvs2[0]), 'Separate dists should match'
    assert not np.array_equal(rvs2[0], rvs2[1]), 'Multiple calls to the same dist should not match'

    return dist, dist2


@sc.timer()
def test_reset(m=m):
    """ Check that reset works as expected """
    sc.heading('Testing reset')

    # Create and draw two sets of random numbers
    dist = ss.random(seed=533, strict=False).init()
    r1 = dist.rvs(m)
    r2 = dist.rvs(m)
    assert all(r1 != r2)

    # Reset to the most recent state
    dist.reset(-1)
    r3 = dist.rvs(m)
    assert all(r3 == r2)

    # Reset to the initial state
    dist.reset(0)
    r4 = dist.rvs(m)
    assert all(r4 == r1)

    for r in [r1, r2, r3, r4]:
        print(r)

    return dist


@sc.timer()
def test_callable(n=n):
    """ Test callable parameters """
    sc.heading('Testing a uniform distribution with callable parameters')

    sim = ss.mock_sim(n_agents=10)

    # Define a parameter as a function
    def custom_loc(module, sim, uids):
        out = sim.people.age[uids]
        return out

    scale = 1
    d1 = ss.normal(name='callable', loc=custom_loc, scale=scale).init(sim=sim)
    d2 = ss.lognorm_ex(name='callable', mean=custom_loc, std=scale).init(sim=sim)

    uids = np.array([1, 3, 7, 9])
    draws1 = d1.rvs(uids)
    draws2 = d2.rvs(uids)
    print(f'Input ages were: {sim.people.age[uids]}')
    print(f'Output samples were: {draws1}, {draws2}')

    for draws in [draws1, draws2]:
        meandiff = np.abs(sim.people.age[uids] - draws).mean()
        assert meandiff < scale*3, 'Outputs should match ages'

    return d1


@sc.timer()
def test_array(n=n):
    """ Test array parameters """
    sc.heading('Testing uniform with a array parameters')

    uids = np.array([1, 3])
    low  = np.array([1, 100]) # Low
    high = np.array([3, 125]) # High

    d = ss.uniform(low=low, high=high, strict=False).init(slots=np.arange(uids.max()+1))
    draws = d.rvs(uids)
    print(f'Uniform sample for uids {uids} returned {draws}')

    assert len(draws) == len(uids)
    for i in range(len(uids)):
        assert low[i] < draws[i] < low[i] + high[i], 'Invalid value'
    return draws


@sc.timer()
def test_repeat_slot():
    """ Test behavior of repeated slots """
    sc.heading('Test behavior of repeated slots')

    # Initialize parameters
    slots = np.array([4,2,3,2,2,3])
    n = len(slots)
    uids = np.arange(n)
    low = np.arange(n)
    high = low + 1

    # Draw values
    d = ss.uniform(low=low, high=high, strict=False).init(slots=slots)
    draws = d.rvs(uids)

    # Print and test
    print(f'Uniform sample for slots {slots} returned {draws}')
    assert len(draws) == len(slots)

    unique_slots = np.unique(slots)
    for s in unique_slots:
        inds = np.where(slots==s)[0]
        frac, integ = np.modf(draws[inds])
        assert np.allclose(integ, low[inds]), 'Integral part should match the low parameter'
        assert np.allclose(frac, frac[0]), 'Same random numbers, so should be same fractional part'
    return draws


def make_mock_modules():
    """ Create mock modules for the tests to use """
    mod = sc.objdict()
    mod.year  = ss.mock_module(dt=ss.year)
    mod.month = ss.mock_module(dt=ss.month)
    mod.week  = ss.mock_module(dt=ss.week)
    return mod


@sc.timer()
def test_timepar_dists():
    """ Test interaction of distributions and timepars """
    sc.heading('Test interaction of distributions and timepars')

    # Set parameters
    n = int(10e4)

    # Mock modules
    mock_mods = make_mock_modules()

    # Create time parameters
    v = sc.objdict()
    v.base = 30.0
    v.dur = ss.years(30)
    v.freq = ss.freqperyear(30)

    rtol = 0.1  # Be somewhat generous with the uncertainty
    check_which = ['linear', 'unitless', 'special']

    if 'linear' in check_which:
        # Check distributions that scale linearly with time, with the parameter we'll set
        print('Testing linear distributions ...')
        linear_dists = dict(
            constant   = 'v',
            uniform    = 'high',
            normal     = 'loc',
            lognorm_ex = 'mean',
            expon      = 'scale',
            weibull    = 'scale',
            gamma      = 'scale',
        )

        for name,par in linear_dists.items():
            dist_class = getattr(ss, name)

            for module in mock_mods.values():

                # Create the dists, the first parameter of which should have time units
                dists = sc.objdict()
                for key,val in v.items():
                    pardict = {par:val} # Convert to a tiny dictionary to insert the correct name
                    dists[key] = dist_class(**pardict, name=key, module=module, strict=False).init()

                # Create the random variates
                rvs = sc.objdict()
                for k,dist in dists.items():
                    rvs[k] = dist.rvs(n)

                # Check that the distributions match
                ratio = ss.years(1)/module.t.dt
                expected = rvs.base.mean()
                expected_dur = expected*ratio
                expected_rate = expected/ratio
                actual_dur = rvs.dur.mean()
                actual_rate = rvs.freq.mean()
                assert np.isclose(expected_dur, actual_dur, rtol=rtol), f'Duration not close for {name}: {expected_dur:n} ≠ {actual_dur:n}'
                assert np.isclose(expected_rate, actual_rate, rtol=rtol), f'Rate not close for {name}: {expected_rate:n} ≠ {actual_rate:n}'
                sc.printgreen(f'✓ {name} passed: {expected_dur:n} ≈ {actual_dur:n} with dt={module.t.dt}')

    if 'unitless' in check_which:
        # Check that unitless distributions fail
        print('Testing unitless distributions ...')
        par = ss.dur(10, 'years')
        unitless_dists = ['lognorm_im', 'randint']
        for name in unitless_dists:
            dist_class = getattr(ss, name)
            with pytest.raises(NotImplementedError):
                dist = dist_class(par, name='notimplemented', module=mock_mods.year, strict=False).init()
                dist.rvs(n)
            sc.printgreen(f'✓ {name} passed: raised appropriate error')

    if 'special' in check_which:
        # Check special distributions
        print('Testing choice distribution ...')
        a = ss.days(np.array([10, 20, 30]))
        p = np.array([0.2, 0.3, 0.5])
        for module in mock_mods.values():
            choice = ss.choice(a=a, p=p, module=module, strict=False).init()
            mean = choice.rvs(n).mean()
            expected = (a*p/module.t.dt).sum()
            assert np.isclose(mean, expected, rtol=rtol), f'Choice values do not match: {mean:n} ≠ {expected:n}'
            sc.printgreen(f'✓ Choice passed: {expected:n} ≈ {mean:n} with dt={module.t.dt}')

        print('Testing Poisson distribution ...')
        lam1 = ss.freqperyear(1)
        lam2 = ss.perday(1)
        for module in mock_mods.values():
            poi1 = ss.poisson(lam=lam1, module=module, strict=False).init()
            poi2 = ss.poisson(lam=lam2, module=module, strict=False).init()
            mean1 = poi1.rvs(n).mean()
            mean2 = poi2.rvs(n).mean()
            expected1 = lam1*module.t.dt
            expected2 = lam2*module.t.dt
            assert np.isclose(mean1, expected1, rtol=rtol), f'Poisson values do not match for {lam1}: {mean1:n} ≠ {expected1:n}'
            assert np.isclose(mean2, expected2, rtol=rtol), f'Poisson values do not match for {lam2}: {mean2:n} ≠ {expected2:n}'
            sc.printgreen(f'✓ Poisson passed: {lam1} {expected1:n} ≈ {mean1:n} with dt={module.t.dt}')
            sc.printgreen(f'✓ Poisson passed: {lam2} {expected2:n} ≈ {mean2:n} with dt={module.t.dt}')

        print('Testing Bernoulli distribution ...')
        p1 = ss.probperday(0.1/365)
        p2 = ss.probperyear(0.1)
        ber1 = ss.bernoulli(p=p1, module=mock_mods.year, strict=False).init()
        ber2 = ss.bernoulli(p=p2, module=mock_mods.year, strict=False).init()
        mean1 = ber1.rvs(n).mean()
        mean2 = ber2.rvs(n).mean()
        assert np.isclose(mean1, mean2, rtol=rtol), f'Bernoulli values do not match for {ber1} and {ber2}: {mean1:n} ≠ {mean2:n}'
        sc.printgreen(f'✓ bernoulli passed: {mean1:n} ≈ {mean2:n}')

    print('Testing different syntaxes for adding timepars')
    kw = dict(dt=ss.days(1))
    mean = 10
    std = 0.1

    d = sc.objdict()
    d.a = ss.normal(ss.years(mean), ss.years(std)).mock(**kw)
    d.b = ss.normal(mean, std, unit=ss.years).mock(**kw)
    d.c = ss.years(ss.normal(mean, std)).mock(**kw)

    n = int(1e6)
    rvs = sc.objdict()
    for key,dist in d.items():
        rvs.key = dist.rvs(n)

    factor = ss.time.factors.years.days
    expected = factor*mean

    means = [rvs.mean() for rvs in rvs.values()]
    assert all([np.isclose(expected, m) for m in means]), f'Expecting timepar scaling to match {expected = } and {means = }'

    return ber2


@sc.timer()
def test_timepar_callable():
    sc.heading('Test that timepars work with (some) callable functions')

    # In this case, the callable returns a time par, which then gets further converted
    print('Testing callable parameters with regular-scaling distributions')

    n = 10_000
    rtol = 0.05

    mock_mods = make_mock_modules()

    def call_scalar(module, sim, uids):
        return ss.years(5.0)

    for module in mock_mods.values():
        d = ss.normal(call_scalar, ss.days(1), module=module, strict=False)
        d.init()
        expected = ss.years(5.0)/module.dt
        actual = d.rvs(n).mean()
        assert np.isclose(actual, expected, rtol=rtol)

    print('Testing callable parameters with Bernoulli distributions')
    n = 1000
    sim = ss.mock_sim(n_agents=n)
    uids = np.random.choice(n, size=n//2, replace=False)
    age = sim.people.age[uids]
    mean = age.mean()
    young = sc.findinds(age<=mean)
    old = sc.findinds(age>mean)
    p_young = 0.1
    p_old = 0.2

    def age_prob(module, sim, uids):
        dt = module.t.dt
        out = np.zeros_like(uids)
        out[young] = ss.peryear(p_young)*dt
        out[old]   = ss.peryear(p_old)*dt
        return out

    sim = ss.mock_sim(n_agents=100_000)
    uids = np.random.choice(n, size=n//2, replace=False)
    age = sim.people.age[uids]
    mean = age.mean()
    young = sc.findinds(age<=mean)
    old = sc.findinds(age>mean)

    ber2 = ss.bernoulli(age_prob, module=mock_mods.year, strict=False).init(sim=sim)
    a = ber2.rvs(uids).mean()

    ber3 = ss.bernoulli(age_prob, module=mock_mods.month, strict=False).init(sim=sim)
    b = ber3.rvs(uids).mean()

    rtol = 0.2  # We're dealing with order-of-magnitude differences but small numbers, so be generous to avoid random failures
    assert np.isclose(a, b*12, rtol=rtol), f'Callable Bernoulli sums did not match: {a} ≠ {b*12}'

    return


# %% Run as a script
if __name__ == '__main__':
    do_plot = True
    sc.options(interactive=do_plot)
    # ss.options.warnings = 'error' # Turn warnings into exceptions for debugging

    T = sc.timer()

    o1 = test_dist()
    o2 = test_custom_dists(do_plot=do_plot)
    o3 = test_dists(do_plot=do_plot)
    o4 = test_scipy()
    o5 = test_exceptions()
    o6 = test_reset()
    o7 = test_callable()
    o8 = test_array()
    o9 = test_repeat_slot()
    o10 = test_timepar_dists()
    o10 = test_timepar_callable()

    T.toc()



import starsim as ss
import numpy as np
import pytest

@pytest.mark.parametrize('age_range,expected', [
    ('5-9', (5, 9)),
    ('5 to 9', (5, 9)),
    ('<5', (0, 5)),
    ('95+', (95, np.inf)),
    ('>95', (95, np.inf)),
])
def test_parse_age_range_valid(age_range, expected):
    assert ss.parse_age_range(age_range) == expected

@pytest.mark.parametrize('age_range', [
    '-1-4',
    '-5',
])
def test_parse_age_range_invalid(age_range):
    with pytest.raises(Exception):
        ss.parse_age_range(age_range)



"""
Test the Loop class
"""

# %% Imports and settings
import sciris as sc
import starsim as ss

sc.options(interactive=False) # Assume not running interactively

pars = sc.objdict(
    dur      = ss.years(20),
    n_agents = 1000,
    diseases = 'sis',
    networks = 'random',
)


@sc.timer()
def test_run_options():
    sc.heading('Testing run options...')
    s1 = ss.Sim(pars).init()
    s2 = ss.Sim(pars).init()

    # Automatic run
    s1.run()

    # Manual run
    while s2.loop.index < len(s2.loop):
        s2.run_one_step()
    s2.finalize()

    assert s1.summary == s2.summary, 'Sims do not match'

    return s2.loop


@sc.timer()
def test_loop_plotting():
    sc.heading('Testing loop plotting...')
    sim = ss.Sim(pars).run()
    sim.loop.plot()
    sim.loop.plot_cpu()
    return sim.loop


# %% Run as a script
if __name__ == '__main__':
    do_plot = True
    sc.options(interactive=do_plot)
    T = sc.timer()

    # Run tests
    l1 = test_run_options()
    l2 = test_loop_plotting()

    T.toc()